//------------------------------------------------------------------------------
//
// fb_load_img.cpp 
//
// Copyright(C) SEIKO EPSON CORPORATION 2011-2014. All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Sample compile:
//    g++ -Wall fb_load_img.cpp pgm.cpp ppm.cpp -o fb_load_img
//
//------------------------------------------------------------------------------

#include <unistd.h>
#include <cassert>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <sys/ioctl.h> 
#include <sys/mman.h>

#include "ppm.h"
#include "pgm.h"

//#define FB	"/dev/fb0"
#define FB	"/dev/fb1"

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;


static u16 ColorToRGB565(unsigned r8, unsigned g8, unsigned b8);
static  u8 ColorTo8bpp(unsigned r8, unsigned g8, unsigned b8);

int main(int argc, char **argv)
{
	int iscolor,fbfd = 0;
	struct fb_var_screeninfo vinfo;
	struct fb_fix_screeninfo finfo;
	long int screensize = 0;
	char *fbp = 0;
	pgm a_pgm;
	ppm a_ppm;
		
	//check for imagefile
	if (argc != 2) {
		printf("Usage: fb_load_img <image.pgm|image.ppm>\n");
		return -1;
	}
	
	//open the framebuffer device
	fbfd = open(FB, O_RDWR);
	if (!fbfd) {
		printf("Error: cannot open framebuffer device \"%s\".\n",FB);
		exit(1);
	}
	printf("The framebuffer device \"%s\" was opened successfully.\n",FB);

	// Get fixed screen information
	if (ioctl(fbfd, FBIOGET_FSCREENINFO, &finfo)) {
		printf("Error reading fixed information.\n");
		exit(2);
	}
	printf("finfo.line_length:%x\n", finfo.line_length);

	// Get variable screen information
	if (ioctl(fbfd, FBIOGET_VSCREENINFO, &vinfo)) {
		printf("Error reading variable information.\n");
		exit(3);
	}
	printf("%dx%d, %dbpp, xoffset:%x, yoffset:%x\n", vinfo.xres, vinfo.yres, vinfo.bits_per_pixel,vinfo.xoffset,vinfo.yoffset);

	// Figure out the size of the screen in bytes
	screensize = vinfo.xres * vinfo.yres * vinfo.bits_per_pixel / 8;

	// Map the device to memory
	fbp = (char *)mmap(0, screensize, PROT_READ | PROT_WRITE, MAP_SHARED, fbfd, 0);
	if ((int)fbp == -1) {
		printf("Error: failed to map framebuffer device to memory.\n");
		exit(4);
	}
	printf("The framebuffer device was mapped to memory successfully. (Addr=%08X)\n",(unsigned)fbp);
	
	//check type of image file, ppm or pgm
	iscolor = (strstr(argv[1],".ppm") != NULL);

	//if ppm image file, process based on framebuffer color depth and write to framebuffer
	if (iscolor) {
		a_ppm.read_file(argv[1]);
		int bc = a_ppm.width() * a_ppm.height();
		
		if (vinfo.bits_per_pixel == 24) {
			u16 *dst = (u16*) fbp;
			printf("Drawing RGB888 image %dx%d..\n", a_ppm.width(),a_ppm.height());
		
			for (int i = 0; i < bc*3; i += 6) {
				*dst++ = (a_ppm.data(i) << 8) | a_ppm.data(i+1);
				*dst++ = (a_ppm.data(i+2) << 8) | a_ppm.data(i+3);
				*dst++ = (a_ppm.data(i+4) << 8) | a_ppm.data(i+5); 
			}
		}
		else if (vinfo.bits_per_pixel == 16) {
			u16 *dst = (u16*) fbp;
			printf("Drawing RGB565 image %dx%d..\n",
				a_ppm.width(),a_ppm.height());
		
			for (int i = 0; i < bc*3; i += 3) {
				*dst++ = ColorToRGB565(a_ppm.data(i),a_ppm.data(i+1),a_ppm.data(i+2));
			}
		} else if (vinfo.bits_per_pixel == 8) {
			u8 *dst = (u8*) fbp;
			printf("Drawing 8bpp image %dx%d..\n",
				a_ppm.width(),a_ppm.height());
		
			for (int i = 0; i < bc*3; i += 3) {
				*dst++ = ColorTo8bpp(a_ppm.data(i),a_ppm.data(i+1),a_ppm.data(i+2));
			}
		} else if (vinfo.bits_per_pixel == 4) {
			u8 *dst = (u8*) fbp;
			printf("Drawing 4BPP image %dx%d..\n",
				a_ppm.width(),a_ppm.height());
		
			for (int i = 0; i < bc*3; i += 6) {
				u8 n1 = ColorTo8bpp(a_ppm.data(i),a_ppm.data(i+1),a_ppm.data(i+2));
				n1 >>= 4;
				u8 n2 = ColorTo8bpp(a_ppm.data(i+3),a_ppm.data(i+4),a_ppm.data(i+5));
				n2 &= 0xF0;
				*dst++ = n2 | n1;
			}
		} else {
			printf(" Unsupported color depth %dbpp\n",vinfo.bits_per_pixel);
		}
	//else if pgm image file, process as grayscale and write to framebuffer if possible
	} else {
		if (strstr(argv[1],".pgm") == NULL) {
			printf("Error: Grayscale image file [.pgm] expected\n");
			goto Done;
		}
		
		a_pgm.read_file(argv[1]);
		int bc = a_pgm.width() * a_pgm.height();
		
		if (vinfo.bits_per_pixel == 8) {
			printf("Drawing grayscale image 8bpp...\n");
			u8 *dst = (u8*) fbp;
			for (int i = 0; i < bc; i++) {
				*dst++ = a_pgm.data(i);
			}
		} else if (vinfo.bits_per_pixel == 4) {
			printf("Drawing grayscale image 4bpp...\n");
			u8 *dst = (u8*) fbp;
			for (int i = 0; i < bc; i+=2) {
				u8 n1 = a_pgm.data(i);
				n1 >>= 4;
				u8 n2 = a_pgm.data(i+1);
				n2 &= 0xF0;
				*dst++ = n2 | n1;
			}
		} else {
			printf(" Unsupported color depth %dbpp\n",vinfo.bits_per_pixel);
		}
	}

Done:
	munmap(fbp, screensize);
	close(fbfd);
	return 0;
}


static u8 ColorTo8bpp(unsigned r8, unsigned g8, unsigned b8)
{
	if (r8 > 255 || b8 > 255 || g8 > 255)
		printf("%x %x %x\n",r8,g8,b8);
		
	unsigned gray = (r8*30 + g8*59 + b8*11)/100;
	gray &= 0xFF;
	return gray; 
}


static u16 ColorToRGB565(unsigned r8, unsigned g8, unsigned b8)
{
	u16 r,g,b;
	b =  (b8>>3) & 0x1f;
	g = ((g8>>2) & 0x3f) << 5;
	r = ((r8>>3) & 0x1f) << 11;
	return r | g | b; 
}
