//------------------------------------------------------------------------------
//
// ppm.cpp 
//
// Copyright(C) SEIKO EPSON CORPORATION 2011-2014. All rights reserved.
//
// Source code derived/ported from work originally done by E Ink.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
//------------------------------------------------------------------------------

#include <cassert>
#include "error.h"
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string.h>

#include "ppm.h"


using namespace std;


ppm::ppm(void) :	_width(0), _height(0), _size(0), _data(0)
{

}

ppm::~ppm(void)
{
	if (_data) 
		delete [] _data;
}

void ppm::init(void)
{
	if (_data) 
		delete [] _data;
		
	_width = 0;
	_height = 0;
	_size = 0;
}

void ppm::read_file(const char * fn)
{
	std::ifstream ifs(fn);
	if (!ifs.is_open()) 
		error(1, 0, "!!! ERROR: failed to open %s for read", fn);

	int state = 0;
	char str[32];
	int v;
	int max_pix_val = 0;

	while (state < 4) {
		ifs >> str;
		
		if (ifs.eof()) 
			error(1, 0, "!!! ERROR: unexpected EOF in %s", fn);
			
		if (str[0] == '#') {
			while (true) {
				char c = ifs.get();
				if (c == '\n') 
					break;
			} 
			continue;
		}
		v = atoi(str);
		switch (state) {
		case 0:
			if (strcmp(str, "P3") && strcmp(str, "P6")) 
				error(1, 0, "!!! ERROR: invalid pgm/ppm format (%s) in %s", str, fn);
			state = 1;
			break;
		case 1:
			_width = v;
			state = 2;
			break;
		case 2:
			_height = v;
			state = 3;
			break;
		case 3:
			max_pix_val = v;
			state = 4;
			break;
		}
	}
 
	if (state != 4) 
		error(1, 0, "!!! ERROR: invalid header in %s", fn);

	if (max_pix_val != 255) 
		error(1, 0, "!!! ERROR: invalid max pixel value (%d) in %s", max_pix_val, fn);

	while (true) {
		char c = ifs.get();
		if (c == '\n') 
			break;
	}

	_size = _width * _height * 3;

	if (_data) 
		delete [] _data;

	_data = new unsigned char [ _size ];
	
	if (_data == NULL) 
		error(1, 0, "!!! ERROR: failed to allocate memory for data");

	ifs.read((char*)_data, _size);
	
	if (ifs.gcount() != _size) 
		error(1, 0, "!!! ERROR: failed to read data from %s", fn);

	ifs.close();
}

unsigned ppm::data(int n) const
{
	if ((n < 0) || (n >= _size)) 
		error(1, 0, "!!! ERROR: invalid data number (%d)", n);
		
	return _data[n];
}


void ppm::write_file(const char * fn)
{
	std::ofstream ofs(fn);
	if (!ofs.is_open()) 
		error(1, 0, "!!! ERROR: failed to open %s for write", fn);

	ofs << "P6" << endl
	<< _width << " "
	<< _height << endl
	<< "255" << endl;
	cout.flush();
	ofs.write((char*)_data, _size);
	ofs.close();
}

void ppm::data(int n, unsigned d)
{
	if (_data == 0) {
		_size = _width * _height * 3;
		
		if (_size <= 0) 
			error(1, 0, "!!! ERROR: invalid width (%d) or height (%d)", _width, _height);
			
		if (_data)
			delete [] _data;
			
		_data = new unsigned char [_size];
		
		if (_data == NULL) 
			error(1, 0, "!!! ERROR: failed to allocate memory for data");
	}
	
	if ((n < 0) || (n >= _size)) 
		error(1, 0, "!!! ERROR: invalid data number (%d)", n);
		
	_data[n] = d;
}

