//-----------------------------------------------------------------------------
// linux/drivers/usb/misc/epson-13u11/s1d13u11.c  
//
// This file provides the routines for the Epson S1D13U11 USB core driver.
// 
// Copyright(C) SEIKO EPSON CORPORATION 2010-2013. All rights reserved.
//
// This driver software is distributed as is, without any warranty of any kind,
// either express or implied as further specified in the GNU Public License. This
// software may be used and distributed according to the terms of the GNU Public
// License, version 2. See the file COPYING in the main directory of this archive
// for more details.
//
// You should have received a copy of the GNU General Public License along with
// this program. If not, see <http://www.gnu.org/licenses/>.
//
//---------------------------------------------------------------------------- 

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/usb.h>
#include <linux/slab.h>
#include <linux/kthread.h>

#include "s1d13u11.h"
#include "s1d13u11_ctl.h"
#include "s1d13u11_fb.h"


// S1D13U11_LCDC_DEF defines
#define	S1D13U11_DEF_INT_INTERVAL		4		/* interval time( High Speed [micro sec] ) */
#define	S1D13U11_DEF_INT_BUF_LEN		128
#define	S1D13U11_DEF_INT_MSG_LEN		S1D13U11_DEF_INT_BUF_LEN
#define	S1D13U11_DEF_BULKIN_TIMEOUT		10000		/* timeout */
#define	S1D13U11_DEF_BULKOUT_TIMEOUT	10000		/* timeout */
#define S1D13U11_DEF_STALL_HALT_OUT		0x01
#define S1D13U11_DEF_STALL_HALT_IN		0x82


// LCDC vendor core structure
struct s1d13u11_vendor {
	struct usb_interface	*intf;
	struct usb_device_id	*devid;
	struct usb_device		*usbdev;
	int						pipebout;
	int						pipebin;
	int						pipeint;
	int						pipedisp;
	struct urb				*urb;
};


spinlock_t			lcdc_lock;			// use demo application
struct semaphore	lcdc_cmdlock;		// use demo application
spinlock_t			lcdc_testint_lock;	// use test tool application

struct semaphore	lcdc_intlock;		// use demo application

static struct s1d13u11_vendor	*lcdc[S1D13U11_DISP_IF_MAX];
static s1d13u11_callback_t 		lcdc_callback[S1D13U11_DEV_ID_MAX][S1D13U11_EV_TYPE_MAX];
static unsigned char 			*lcdc_buf = NULL;

static struct	task_struct *wup_th;
static int		flag_th=0;
static atomic_t total_wupthreads = ATOMIC_INIT(0);

static unsigned short	s_resumeData[] = {
	0x006E,	0x0001,	// General Purpose Output Register 1 (Power on CASIO Panel)
	0xFFFF,	0x0080,	// 100ms wait
	0x0004,	0x0081,	// PLL Control Register
	0xFFFF,	0x0001,	// 1ms wait
	0x0012,	0x00C2,	// Clock Source Select Register(PCLK=9MHzASS=SDCLK)
	0x0068,	0x00E8,	// Power Save Register
	0x0068,	0x0000,	// Power Save Register
	0x0068,	0x0001,	// Power Save Register
	0xFFFF,	0x0001,	// 1ms wait
	0x0084,	0x0000,	// SDRAM self-refresh off
	0xFFFF,	0x0001,	// 1ms wait
	0x002A,	0x0001,	//
	0x0050,	0x0080,	//
	0x006E,	0x0000,	// General Purpose Output Register 1 (STBY OFF)
	0xFFFF,	0x0001,	// 1ms wait
	0x0070,	0x0086	// PWM Control Register(PWM On)
} ;

static unsigned short	s_suspendData[] = {
	0x0070,	0x0084,	// PWM Control Register(PWM Off)
	0x0070,	0x0000,	// PWM Control Register(PWM Off)
	0x006E,	0x0001,	// General Purpose Output Register 1 (STBY ON)
	0xFFFF,	0x0080,	// 100ms wait
	0x002A,	0x0000,	// Display off
	0x009E,	0x0000,	// Alpha-blending disable
	0xFFFF,	0x0001,	// 1ms wait
	0x0084,	0x0008,	// SDRAM self-refresh
	0xFFFF,	0x0001,	// 1ms wait
	0x0068,	0x0000,	// Power Save Register
	0x0012,	0x0001,	// Clock Source Select Register
	0x0004,	0x0001,	// PLL Control Register
	0xFFFF,	0x001E,	// 30ms wait
	0x006E,	0x0003	// General Purpose Output Register 1 (Power off CASIO Panel)
} ;


#ifdef CONFIG_S1D13U11_DEBUG_MESSAGES
static void dump_data(char *msg, unsigned char *data, int len);
#endif

// prototypes
static void s1d13u11_cmp_int(struct urb *urb, struct pt_regs *regs);


int	debugLockCntInt = 0 ;
static int suspendFlg = 0 ;
static DECLARE_WAIT_QUEUE_HEAD(wqueue);
static int wflag=0;


//static struct timer_list sustlist;
wait_queue_head_t  g_waitq;

static struct timer_list	resumeTimer;

#define LCDC_SEQUENCE_DELAY		0x00
#define LCDC_SEQUENCE_IRQ		0x01
#define	LCDC_SEQUENCE_NONE		0x02

#define	LCDC_SEQUENCE_WRITE		0x00
#define	LCDC_SEQUENCE_READ		0x01
#define	LCDC_SEQUENCE_CSENB		0x02
#define	LCDC_SEQUENCE_CSDIS		0x03
#define	LCDC_SEQUENCE_INTDIS	0x04
#define	LCDC_SEQUENCE_INTDNA	0x05
#define	LCDC_SEQUENCE_WAIT		0x06

static int s_timeCnt = 0 ;
static void resume_timer_handler(unsigned long data);
static int s1d13u11_remotewup_thr(void* data);


static void s1d13u11_start_timer( void )
{
	init_timer( &resumeTimer );

	resumeTimer.expires  = jiffies + 500;
	resumeTimer.data     = ( unsigned long )jiffies;
	resumeTimer.function = resume_timer_handler;

	add_timer( &resumeTimer );

	return ;
}


static void resume_timer_handler(unsigned long data)
{
	if( s_timeCnt < 2 ) {
		s_timeCnt++ ;
		s1d13u11_start_timer();
		return ;
	}
	wake_up_process(wup_th);
	return ;
}


//---------------------------------------------------------------------------- 
// s1d13u11_remotewup_thr()
// Remote Wakeup Notith
// - Loop. data not zero.
//
// param	data	:	data
//
// retval  	0
//   if error, then error code
//
//---------------------------------------------------------------------------- 
static int s1d13u11_remotewup_thr(void* data)
{

	DPRINT( "Int Restart Start!!(%s-%d)\n",__func__,__LINE__ );

//	wait_event_interruptible(wqueue, (wflag!=0));	/* pgr695 */
//	wflag=0;

	printk( "Int Restart Wait...Restart!!(%s-%d)\n",__func__,__LINE__ );

	printk( "Int Restart Start Sleep!!(%s-%d)\n",__func__,__LINE__ );
//	lcdc[S1D13U11_DISP_IF_0]->usbdev->dev.power.power_state.event = PM_EVENT_ON;
//	lcdc[S1D13U11_DISP_IF_0]->usbdev->bus->controller->power.power_state.event = PM_EVENT_FREEZE ;
//	lcdc[S1D13U11_DISP_IF_0]->usbdev->state = USB_STATE_ATTACHED ;

	printk( "Int Restart Start!!(%s-%d)\n",__func__,__LINE__ );

#if 0
	ret = s1d13u11_core_resume( lcdc[S1D13U11_DISP_IF_0]->usbdev );
	if(ret!=0) {
		printk("s1d13u11_core_resume %s-%d ERROR[%d] \n", __FUNCTION__,__LINE__,ret );
	}
#endif

#if 0
	// Restart Interrupt
	ret = s1d13u11_send_int(
			UCHARIS(S1D13U11_DISP_IF_0),
			S1D13U11_DEF_INT_MSG_LEN,
			lcdc_buf );
	if(ret){
		printk("%s ERROR %d \n", __FUNCTION__, ret);
	}
#endif

//	ret = usb_lock_device_for_reset(lcdc[S1D13U11_DISP_IF_0]->usbdev, lcdc[S1D13U11_DISP_IF_0]->intf);

	// Restart Interrupt
//	ret = usb_submit_urb(lcdc[S1D13U11_DISP_IF_0]->urb, SLAB_ATOMIC);
//	if(ret){
//		printk("%s ERROR %d \n", __FUNCTION__, ret);
//	}

	printk( "Int Restart Stop!!(%s-%d)\n",__func__,__LINE__ );

	wflag = 0 ;
	return 0 ;
}


//---------------------------------------------------------------------------- 
// s1d13u11_send_cbw()
// Send "CBW" Message
// - Check input parameter.
// - Send bulk message synchronously.
// - Send bulk error check.
//   - case STALL(-EPIPE)
//     - "s1d13u11_clear_halt" call.
//     - get error message.
//   - case Not success
//     - Error return .
//
// param[in]	ifno	: Interface number(always 0)
// param[in]	len		: Send message length.
// param[in]	buf		: Send message data.
// param[out]	rcvlen	: Receive message length.
// param[out]	rcvbuf	: Recvive message data.
//
// retval  	0		: success
// retval  	not 0	: if error, then error code
//
//---------------------------------------------------------------------------- 
int s1d13u11_send_cbw(
			unsigned char	ifno,
			ssize_t		len,
			unsigned char	*buf,
			ssize_t		rcvlen,
			unsigned char	*rcvbuf)
{
	int	wlen=0;
	int	ret=0;
	int	rlen=0;			/* actual read */

	DPRINT("%s len %d \n", __FUNCTION__, len);

	/* check parameter */
	if(ifno>UCHARIS(S1D13U11_DISP_IF_1)){
		return -EINVAL;
	}
	if(lcdc[ifno]==NULL){
		printk("%s ERROR no device \n", __FUNCTION__);
		return -ENODEV;
	}

#ifdef CONFIG_S1D13U11_DEBUG_MESSAGES
	dump_data("send CBW", buf, (int)len );
#endif

	/* send bulk synchronously */
	ret=usb_bulk_msg(
		lcdc[ifno]->usbdev,
		usb_sndbulkpipe(lcdc[ifno]->usbdev, (unsigned int)lcdc[ifno]->pipebout) /* pgr0787 */,
		buf,
		len,
		&wlen,
		S1D13U11_DEF_BULKOUT_TIMEOUT);

	/* STALL Check */
	if( ret == -EPIPE ) {
		/* Clear STALL */
		ret = s1d13u11_clear_halt( UCHARIS(S1D13U11_DISP_IF_0) );
		if( ret==0 ) {
			/* Error Info Get */
			ret=usb_bulk_msg(
				lcdc[ifno]->usbdev,
				usb_rcvbulkpipe(lcdc[ifno]->usbdev, (unsigned int)lcdc[ifno]->pipebin) /* pgr0787 */,
				rcvbuf,
				rcvlen,
				&rlen,
				S1D13U11_DEF_BULKIN_TIMEOUT);
			if( ret!=0 ) {
				printk("%s RECOVER ERROR %d %d %d \n", __FUNCTION__, ret, rcvlen, rlen);
				return -EFAULT ;
			}

#ifdef CONFIG_S1D13U11_DEBUG_MESSAGES
	dump_data("recv CSW", rcvbuf, (int)rlen );
#endif
		}
		return rlen ;
	}

	if((ret!=0)||(len!=wlen)){
		printk("%s LEN ERROR %d %d %d \n", __FUNCTION__, ret, len, wlen);
		return ret ;
	}

	return ret;
}


//---------------------------------------------------------------------------- 
// s1d13u11_recv_csw()
// Receive "CSW" Message
// - Check input parameter.
// - Receive bulk message synchronously.
// - Receive bulk error check.
//   - case STALL(-EPIPE)
//	   - "s1d13u11_clear_halt" call.
//	   - get error message.
//	 - case Not success
//	   - Error return .
//
// param[in]	ifno	: Interface number(always 0)
// param[out]	len		: Receive message length.
// param[out]	buf		: Recvive message data.
//
// retval  	0		: success
// retval  	not 0	: if error, then error code
//
//---------------------------------------------------------------------------- 
int s1d13u11_recv_csw(unsigned char ifno, ssize_t len, unsigned char *buf)
{
	int	rlen=0;			/* actual read */
	int	ret=0;

	DPRINT("%s len %d \n", __FUNCTION__, len);

	/* check parameter */
	if(ifno>UCHARIS(S1D13U11_DISP_IF_1)){
		return -EINVAL;
	}
	if(lcdc[ifno]==NULL){
		printk("%s ERROR no device \n", __FUNCTION__);
		return -ENODEV;
	}

	/* receive bulk synchronously */
	ret=usb_bulk_msg(
		lcdc[ifno]->usbdev,
		usb_rcvbulkpipe(lcdc[ifno]->usbdev, (unsigned int)lcdc[ifno]->pipebin) /* pgr0787 */,
		buf,
		len,
		&rlen,
		S1D13U11_DEF_BULKIN_TIMEOUT);

	/* STALL Check */
	if( ret == -EPIPE ) {
		/* Clear STALL */
		ret = s1d13u11_clear_halt( UCHARIS(S1D13U11_DISP_IF_0) );
		if( ret==0 ) {
			/* Error Info Get */
			ret=usb_bulk_msg(
				lcdc[ifno]->usbdev,
				usb_rcvbulkpipe(lcdc[ifno]->usbdev, (unsigned int)lcdc[ifno]->pipebin) /* pgr0787 */,
				buf,
				len,
				&rlen,
				S1D13U11_DEF_BULKIN_TIMEOUT);
			if( ret!=0 ) {
				DPRINT("%s RECOVER ERROR %d %d %d \n", __FUNCTION__, ret, len, rlen);
				return -EFAULT ;
			}
		}
	}

	if((ret!=0)||(len<rlen)){
		DPRINT("%s LEN ERROR %d %d %d \n", __FUNCTION__, ret, len, rlen);
		return -EFAULT;
	}

	ret = rlen ;

#ifdef CONFIG_S1D13U11_DEBUG_MESSAGES
	dump_data("recv CSW", buf, (int)rlen );
#endif

	return ret;
}

//---------------------------------------------------------------------------- 
// s1d13u11_send_data()
// Send Data Message(Mainly Framebuffer)
// - Check input parameter.
// - Send bulk message synchronously.
// - Send bulk error check.
//   - case Not success
//   - Error return .
//
// param[in]	ifno	: Interface number(always 0)
// param[in]	len		: Send data length.
// param[in]	buf		: Send data data.
//
// retval  	0		: success
// retval  	not 0	: if error, then error code
//
//---------------------------------------------------------------------------- 
int s1d13u11_send_data(unsigned char ifno, ssize_t len, unsigned char *buf)
{
	int	wlen = 0;
	int	ret  = 0;

	DPRINT("%s len %d \n", __FUNCTION__, len);

	/* check parameter */
	if(ifno>UCHARIS(S1D13U11_DISP_IF_0)){
		return -EINVAL;
	}
	if(lcdc[ifno]==NULL){
		printk("%s ERROR no device \n", __FUNCTION__);
		return -ENODEV;
	}

#ifdef CONFIG_S1D13U11_DEBUG_MESSAGES
	dump_data("send data", buf, (int)len );
#endif

	/* send bulk synchronously */
	ret=usb_bulk_msg(
		lcdc[ifno]->usbdev,
		usb_sndbulkpipe(lcdc[ifno]->usbdev, (unsigned int)lcdc[ifno]->pipedisp) /* pgr0787 */,
		buf,
		len,
		&wlen,
		S1D13U11_DEF_BULKOUT_TIMEOUT);
		
	DPRINT("usb_bulk_msg returns: ret=%x,len=%x,wlen=%x\n", ret, len, wlen);

	if((ret!=0)||(len!=wlen)){
		DPRINT("%s ERROR %d %d %d \n", __FUNCTION__, ret, len, wlen);
		return -EFAULT;
	}

	return ret;
}

//---------------------------------------------------------------------------- 
// s1d13u11_recv_data()
// Receive Data Message(Not use DEMO)
// - Receive bulk message synchronously.
//
// param[in]	ifno	: Interface number(always 0)
// param[out]	len		: Receive data length.
// param[out]	buf		: Receive data data.
//
// return  			: actual read length
//
//---------------------------------------------------------------------------- 
int s1d13u11_recv_data(unsigned char ifno, ssize_t len, unsigned char *buf)
{
	DPRINT("%s \n", __FUNCTION__);

	return s1d13u11_recv_csw(ifno, len, buf);
}


//---------------------------------------------------------------------------- 
// s1d13u11_send_int()
// Submit INT transfer.
// - Create urb.
// - Initialize a interrupt urb.
// - Issue an asynchronous transfer request for an endpoint.
// 
// param[in]	ifno	: Interface number(always 0)
// param[in]	len		: Send data length.
// param[in]	buf		: Send data data.
//
// retval		0		: complete
// retval  	not 0	: if error, then error code
//
//---------------------------------------------------------------------------- 
int s1d13u11_send_int(unsigned char ifno, ssize_t len, unsigned char *buf)
{
	struct urb *urb;
	int	ret=0;

	DPRINT("%s len %d \n", __FUNCTION__, len);

	if(lcdc[ifno]==NULL){
		printk("%s ERROR no device \n", __FUNCTION__);
		return -ENODEV;
	}

	/* create urb */
	urb = usb_alloc_urb(0, GFP_KERNEL);
	if( urb==(struct urb*)NULL ){
		return -ENOMEM;
	}
	lcdc[ifno]->urb = urb;

	usb_fill_int_urb(
		urb,
		lcdc[ifno]->usbdev,
		usb_rcvintpipe(	lcdc[ifno]->usbdev,
				(unsigned int)lcdc[ifno]->pipeint) /* pgr0787 */,
		buf,
		len,
		(usb_complete_t)s1d13u11_cmp_int,
		lcdc[ifno],
		S1D13U11_DEF_INT_INTERVAL );

	/* send */
	ret=usb_submit_urb(urb, GFP_KERNEL);
	if(ret){
		DPRINT("%s ERROR %d \n", __FUNCTION__, ret);
	}

	return ret;
}


//---------------------------------------------------------------------------- 
// s1d13u11_cmp_int()
// Complete Interrupt transfer.
// - Check input parameter.
// - Status check.
//   - case UNLINK(ECONNRESET/ENOENT/ESHUTDOWN)
//     - return
// - Event code check.
//   - case LCDC_EVENT
//     - Get event data(LCDC_READ)
//     - Clear event register(LCDC_WRITE)
//     - Interrupt Driver callback.
//   - case SPI_INT_EVENT
//     - Get event data(SPI_ACCESS)
//     - Touch screen Driver callback.
//     - Restart SPI int(EVENT_INT_CONTROL)
//   - case SPI_SEQ_EVENT
//     - Touch screen Driver callback.
//   - case WAKEUP_EVENT
//     - Wakeup Driver callback.
//   - case KEYSCAN_EVENT
//     - Key input Driver callback.
//     - Issue an asynchronous transfer request for an endpoint.
// 
// param[in]	urb		: struct of urb
// param[in]	regs	: struct of pt_regs
//
// retval		0		: complete
// retval  	not 0	: if error, then error code
// 
//---------------------------------------------------------------------------- 
static void s1d13u11_cmp_int(struct urb *urb, struct pt_regs *regs)
{
	struct s1d13u11_vendor		*lcdc;
	struct s1d13u11_disp_event	*devent;
	struct s1d13u11_disp_event	eventStr;
	struct s1d13u11_disp_event_test *etest;
	int	stat=0;

	DPRINT("%s \n", __FUNCTION__);

	/* check parameter */
	if( urb == NULL ){
		printk("%s NO URB \n", __FUNCTION__);
		return;
	}

	if( debugLockCntInt > 0 ) {
		usb_submit_urb(urb, /*SLAB_ATOMIC*/ GFP_ATOMIC);
		printk("Continue!!(%s-%d)\n",__func__,__LINE__);
		return ;
	}

	switch( urb->status ) {
	    case 0 :	// SUCCESS
		break ;
	    case -ECONNRESET :	// UNLINK!!
	    case -ENOENT :
	    case -ESHUTDOWN :
		// Disconnect Notice
		memset( (void *)&eventStr, 0x00, sizeof(struct s1d13u11_disp_event) );

		eventStr.etype = S1D13U11_DISP_EVENT_INT_STAT_DISCONNECT ;

		if(lcdc_callback[S1D13U11_DEV_ID_INT][S1D13U11_EV_TYPE_INT] != (s1d13u11_callback_t)NULL) {
			lcdc_callback[S1D13U11_DEV_ID_INT][S1D13U11_EV_TYPE_INT](
							(unsigned int)sizeof(struct s1d13u11_disp_event),
							(unsigned char *)&eventStr );
		}

		return ;
	    default :
		stat=usb_submit_urb(urb, /*SLAB_ATOMIC*/ GFP_ATOMIC);
		if(stat){
			/* error on resubmit */
			printk("%s ERROR in resubmit %d \n", __FUNCTION__, stat);
		}
		return ;
	}

	lcdc=(struct s1d13u11_vendor*)urb->context;

#ifdef CONFIG_S1D13U11_DEBUG_MESSAGES
	dump_data("recv INT", urb->transfer_buffer, urb->actual_length);
#endif

	/* check event */
	devent=(struct s1d13u11_disp_event*)urb->transfer_buffer;

	if( !devent ) {
		// Event error
		/* send again */
		stat = usb_submit_urb(urb, GFP_ATOMIC);
		if(stat){
			/* error on resubmit */
			printk("%s ERROR in resubmit %d \n", __FUNCTION__, stat);
		}
		return ;
	}

	etest=(struct s1d13u11_disp_event_test *)urb->transfer_buffer;

	DPRINT( "devent->etype[0x%X](%s-%d)\n",devent->etype,__func__,__LINE__ );

	if(devent->etype == (unsigned char)S1D13U11_DISP_EVENT_INT_STAT_LCDC){
		;
	}
	else if(devent->etype == (unsigned char)S1D13U11_DISP_EVENT_INT_STAT_SPIINT) {
		// SPI INTERRUPT NOTICE
		if(lcdc_callback[S1D13U11_DEV_ID_TS][S1D13U11_EV_TYPE_INT] != (s1d13u11_callback_t)NULL) {

			lcdc_callback[S1D13U11_DEV_ID_TS][S1D13U11_EV_TYPE_INT](
							(unsigned int)urb->actual_length,
							(unsigned char *)urb->transfer_buffer );
		}
	}
	else if(devent->etype == (unsigned char)S1D13U11_DISP_EVENT_INT_STAT_SPISEQ) {

		// SPI SEQUENCE NOTICE
		if(lcdc_callback[S1D13U11_DEV_ID_TS][S1D13U11_EV_TYPE_SEQ] != (s1d13u11_callback_t)NULL) {

			lcdc_callback[S1D13U11_DEV_ID_TS][S1D13U11_EV_TYPE_SEQ](
							(unsigned int)urb->actual_length,
							(unsigned char *)urb->transfer_buffer );
		}
	}
	else if(devent->etype == (unsigned char)S1D13U11_DISP_EVENT_INT_STAT_GPI) {
		// LCDC INTERRUPT NOTICE(DEMO Not Use!)
		if(lcdc_callback[S1D13U11_DEV_ID_FB][S1D13U11_EV_TYPE_INT] != (s1d13u11_callback_t)NULL) {

			lcdc_callback[S1D13U11_DEV_ID_FB][S1D13U11_EV_TYPE_INT](
						(unsigned int)urb->actual_length,
						(unsigned char *)urb->transfer_buffer );
		}
	}

	else {
		// error on unknow event type
		DPRINT("%s ERROR event type[%d]\n", __FUNCTION__, devent->etype);
	}

	/* send again */
	stat = usb_submit_urb(urb, GFP_ATOMIC);
	if(stat){
		/* error on resubmit */
		printk("%s ERROR in resubmit %d \n", __FUNCTION__, stat);
	}

	return ;
}


//---------------------------------------------------------------------------- 
// s1d13u11_clear_halt()
// Halt clear.
// - Message is send back out a sending end point' to CLEAR FEATURE'.
// - Toggle of the sending end point is cleared.
// - Message is send back out a receive end point' to CLEAR FEATURE'.
// - Toggle of the receive end point is cleared.
//
// param[in]	ifno	: Interface number(always 0)
//
// retval		0		: success
// retval  	not 0	: if error, then error code
//
//---------------------------------------------------------------------------- 
int s1d13u11_clear_halt(unsigned char ifno )
{
	int	ret ;
	int spipe, rpipe;
	int endp = usb_pipeendpoint( (unsigned int)lcdc[ifno]->pipebout );
	
	DPRINT("%s \n", __FUNCTION__);

	if(usb_pipein(ifno)) {
		endp |= USB_DIR_IN;
	}

	// Reset Recovery
	ret = usb_control_msg(
			lcdc[ifno]->usbdev,
			usb_sndctrlpipe(lcdc[ifno]->usbdev, 0), /* pgr0787 */
			USB_REQ_CLEAR_FEATURE,
			USB_RECIP_ENDPOINT,
			USB_ENDPOINT_HALT,
			S1D13U11_DEF_STALL_HALT_OUT,
			NULL, 0,
			USB_CTRL_SET_TIMEOUT);

	if(ret!=0) {
		DPRINT("%s-%d ERROR[%d] \n", __FUNCTION__,__LINE__,ret );
		return ret ;
	}

	/* toggle was reset by the clear */
	spipe = usb_sndbulkpipe(
				lcdc[ifno]->usbdev,
				(unsigned int)lcdc[ifno]->pipebout) ;
//	usb_settoggle(
//			lcdc[ifno]->usbdev,
//			usb_pipeendpoint( spipe ),
//			usb_pipeout( spipe ),
//			0 );

	ret = usb_control_msg(
			lcdc[ifno]->usbdev,
			usb_rcvctrlpipe(lcdc[ifno]->usbdev, 0), /* pgr0787 */
			USB_REQ_CLEAR_FEATURE,
			USB_RECIP_ENDPOINT,
			USB_ENDPOINT_HALT,
			S1D13U11_DEF_STALL_HALT_IN,
			NULL, 0,
			USB_CTRL_SET_TIMEOUT);

	if(ret!=0) {
		DPRINT("%s-%d ERROR[%d] \n", __FUNCTION__,__LINE__,ret );
		return ret ;
	}

	/* toggle was reset by the clear */
	rpipe = usb_rcvbulkpipe(
				lcdc[ifno]->usbdev,
				(unsigned int)lcdc[ifno]->pipebin) ;
//	usb_settoggle(
//			lcdc[ifno]->usbdev,
//			usb_pipeendpoint( rpipe ),
//			usb_pipeout( rpipe ),
//			0 );
	return 0 ;
}


//---------------------------------------------------------------------------- 
// s1d13u11_soft_reset()
// Soft reset transfer
// - Send Control message.
//
// param[in]	size	: send length
// param[in]	buf		: send data
//
// retval		0		: success
// retval  	not 0	: if error, then error code
// 
//---------------------------------------------------------------------------- 
int s1d13u11_soft_reset(unsigned int size, unsigned char *buf )
{
	struct s1d13u11_softreset *test_softreset =(struct s1d13u11_softreset *)buf ;
	int	ret = 0 ;

	if(lcdc[S1D13U11_DISP_IF_0]==NULL){
		printk("%s ERROR no device \n", __FUNCTION__);
		return -ENODEV;
	}

	DPRINT("%s \n", __FUNCTION__);

	ret = usb_control_msg(
			lcdc[S1D13U11_DISP_IF_0]->usbdev,	// device
			usb_sndctrlpipe(lcdc[S1D13U11_DISP_IF_0]->usbdev, 0) /* pgr0787 */,
			test_softreset->bRequest,
			test_softreset->bmRequestType,
			test_softreset->wValue,
			test_softreset->wIndex,
			test_softreset->Data,
			test_softreset->wLength,
			S1D13U11_DEF_BULKOUT_TIMEOUT );
	return ret ;
}


//---------------------------------------------------------------------------- 
// s1d13u11_busreset()
// Bus reset transfer
// - "usb_reset_device" call.
//
// param[in]	ifno	: Interface number(always 0)
//
// retval		0		: success
// retval  	not 0	: if error, then error code
// 
//---------------------------------------------------------------------------- 
int s1d13u11_busreset( unsigned char ifno )
{
	int	ret ;

	DPRINT("%s \n", __FUNCTION__);

	if(lcdc[S1D13U11_DISP_IF_0]==NULL){
		printk("%s ERROR no device \n", __FUNCTION__);
		return -ENODEV;
	}

	ret = usb_reset_device( lcdc[S1D13U11_DISP_IF_0]->usbdev );

	// Restart Interrupt
	ret = s1d13u11_send_int(
			UCHARIS(S1D13U11_DISP_IF_0),
			S1D13U11_DEF_INT_MSG_LEN,
			lcdc_buf );
	if(ret){
		DPRINT("%s ERROR %d \n", __FUNCTION__, ret);
	}
	if( lcdc[S1D13U11_DISP_IF_0]->usbdev->dev.power.power_state.event != PM_EVENT_ON ) {
		lcdc[S1D13U11_DISP_IF_0]->usbdev->dev.power.power_state.event = PM_EVENT_ON;
	}

	DPRINT("%s ret[%d] \n", __FUNCTION__,ret );
	return 0 ;
}


//---------------------------------------------------------------------------- 
// s1d13u11_suspend_status()
// Get suspend status.
// - Get power status.
//
// retval	S1D13U11_LCDC_STATUS_RUN		:	Resume
// retval	S1D13U11_LCDC_STATUS_SUSPEND	:	Suspend
//
//---------------------------------------------------------------------------- 
int s1d13u11_suspend_status( void )
{

	if( lcdc[S1D13U11_DISP_IF_0]->usbdev->dev.power.power_state.event == PM_EVENT_ON ) {
		return( S1D13U11_LCDC_STATUS_RUN );
	}

	return( S1D13U11_LCDC_STATUS_SUSPEND );
}


//---------------------------------------------------------------------------- 
// s1d13u11_suspend()
// USB Suspend.
// - Check power status.
// - Send control mesage.
// - "s1d13u11_core_suspend" call.
// - Set power status
//
// param[in]	ifno	: Interface number(always 0)
//
// retval		0		: success
// retval  	not 0	: if error, then error code
// 
//---------------------------------------------------------------------------- 
int s1d13u11_suspend( unsigned char ifno )
{
	int	ret ;
	int	status;
	u16	devstatus;

	DPRINT("%s \n", __FUNCTION__);

	if(lcdc[S1D13U11_DISP_IF_0]==NULL){
		printk("%s ERROR no device \n", __FUNCTION__);
		return -ENODEV;
	}

	//-------------------------------------------------------------
	// create thread
	//-------------------------------------------------------------

	DPRINT( "kthread_create(%s-%d)\n",__func__,__LINE__ );

	wup_th = kthread_create(
			s1d13u11_remotewup_thr, &flag_th, "s1d13u11fb");
	if( IS_ERR(wup_th) ) {
		printk("%s CANNOT create thead \n", __FUNCTION__);
		return( -EFAULT );
	}
	atomic_inc(&total_wupthreads);


	if( lcdc[S1D13U11_DISP_IF_0]->usbdev->dev.power.power_state.event == PM_EVENT_ON ) {
		DPRINT("(%s-%d)\n", __FUNCTION__,__LINE__);

		ret = usb_control_msg(
			lcdc[S1D13U11_DISP_IF_0]->usbdev,
			usb_sndctrlpipe(lcdc[S1D13U11_DISP_IF_0]->usbdev, 0),
			USB_REQ_SET_FEATURE,
			USB_RECIP_DEVICE,
			USB_DEVICE_REMOTE_WAKEUP,
			0,
			NULL, 0,
			USB_CTRL_SET_TIMEOUT);

		DPRINT("ret[%d](%s-%d)\n", ret ,__FUNCTION__,__LINE__);

		if(ret!=0) {
			DPRINT("%s-%d ERROR[%d] \n", __FUNCTION__,__LINE__,ret );
			return ret ;
		}

		status = usb_get_status(lcdc[S1D13U11_DISP_IF_0]->usbdev, USB_RECIP_DEVICE, 0, &devstatus) ;

		DPRINT( "status[0x%X](%s-%d)\n",status,__func__,__LINE__ );

		// s1d13u11_core_suspend( lcdc[S1D13U11_DISP_IF_0]->usbdev );
		ret = 0; 
		if(ret!=0) {
			printk("%s-%d ERROR[%d] \n", __FUNCTION__,__LINE__,ret );
			return ret ;
		}

		status = usb_get_status(lcdc[S1D13U11_DISP_IF_0]->usbdev, USB_RECIP_DEVICE, 0, &devstatus) ;
		DPRINT( "status[0x%X](%s-%d)\n",status,__func__,__LINE__ );

		lcdc[S1D13U11_DISP_IF_0]->usbdev->dev.power.power_state.event = PM_EVENT_SUSPEND ;
	}
	else {
		return( -EFAULT );
	}

	suspendFlg = 1 ;
	return( 0 );
}


//---------------------------------------------------------------------------- 
// s1d13u11_resume()
// USB resume.
// - Check power status.
// - "s1d13u11_core_resume" call.
// - Set power status
// 
// param[in]	ifno	: Interface number(always 0)
//
// retval		0		: success
// retval  	not 0	: if error, then error code
//
//---------------------------------------------------------------------------- 
int s1d13u11_resume( unsigned char ifno )
{
	int	ret ;

	DPRINT("%s \n", __FUNCTION__);

	if(lcdc[S1D13U11_DISP_IF_0]==NULL){
		printk("%s ERROR no device \n", __FUNCTION__);
		return -ENODEV;
	}
	if( lcdc[S1D13U11_DISP_IF_0]->usbdev->dev.power.power_state.event != PM_EVENT_ON ) {
		// s1d13u11_core_resume( lcdc[S1D13U11_DISP_IF_0]->usbdev );
		ret = 0; 
		if(ret!=0) {
			DPRINT("%s-%d ERROR[%d] \n", __FUNCTION__,__LINE__,ret );
			return ret ;
		}

		// Restart Interrupt
		ret = s1d13u11_send_int(
				UCHARIS(S1D13U11_DISP_IF_0),
				S1D13U11_DEF_INT_MSG_LEN,
				lcdc_buf );
		if(ret){
			DPRINT("%s ERROR %d \n", __FUNCTION__, ret);
		}
		if( lcdc[S1D13U11_DISP_IF_0]->usbdev->dev.power.power_state.event != PM_EVENT_ON ) {
			lcdc[S1D13U11_DISP_IF_0]->usbdev->dev.power.power_state.event = PM_EVENT_ON;
		}

	}
	else {
		DPRINT("%s ERROR %d \n", __FUNCTION__, __LINE__);
		return( -EFAULT );
	}
	return( 0 );
}


//---------------------------------------------------------------------------- 
// s1d13u11_recv_int()
// Interrupt callback Registration
// - Check power status.
// - "s1d13u11_core_resume" call.
// - Set power status
// 
// param[in]	id		: driver id
// param[in]	etype	: event type
// param[in]	callback: callback function
//
// retval		0		: success
// retval  	not 0	: if error, then error code
//
//---------------------------------------------------------------------------- 
int s1d13u11_recv_int(unsigned char id, unsigned int etype, s1d13u11_callback_t *callback)
{
	int	ret=0;

	//DPRINT("%s id %d type %d \n", __FUNCTION__, id, etype);

	/* check parameter */
	if( (id>=UCHARIS(S1D13U11_DEV_ID_MAX))
			|| (etype>=(unsigned int)S1D13U11_EV_TYPE_MAX))
	{
		return -EINVAL;
	}

	// register callback
	// if callback is NULL, then deregister
	lcdc_callback[id][etype]=(void*)callback;

	return ret;
}


//---------------------------------------------------------------------------- 
// s1d13u11_get_acctype()
// Get access type
// - "DisInc" parameter get.
//
// param[in]	disinc	: disable incriment
// param[in]	bpp		: bit per pixel
//
// retval				: convert data
// 
//---------------------------------------------------------------------------- 
unsigned char s1d13u11_get_acctype(unsigned char disinc, unsigned char bpp)
{
	unsigned char	cnv;

	cnv = (unsigned char)S1D13U11_DISP_SEL_ENINC;

	switch(bpp){
	case 8:		cnv|=(unsigned char)S1D13U11_DISP_SEL_ACCTYPE_8;	break;
	case 32:	cnv|=(unsigned char)S1D13U11_DISP_SEL_ACCTYPE_32;	break;
	case 16:
	default:	cnv|=(unsigned char)S1D13U11_DISP_SEL_ACCTYPE_16;
	}

	return cnv;
}


//---------------------------------------------------------------------------- 
// s1d13u11_cnv_size()
// convert 4 to 3 byte
// - convert 4 to 3 byte
//
// param[in]	pcvt	: converted
// param[in]	raw		: raw data
//
// retval				: None.
//
//---------------------------------------------------------------------------- 
void s1d13u11_cnv_size(unsigned char *pcvt, unsigned int raw)
{
	unsigned char *p;

	p=(unsigned char*)&raw;
	if(pcvt){
		pcvt[0]=p[0];
		pcvt[1]=p[1];
	}
	return ;
}


//---------------------------------------------------------------------------- 
// s1d13u11_init_peripheral()
// Initialize peripheral driver
// - Start INT transfer
// 
// retval		0		: success
// retval  	not 0	: if error, then error code
// 
//---------------------------------------------------------------------------- 
static int s1d13u11_init_peripheral(void)
{
	int	ret=0;

	/* start INT transfer */
	ret=s1d13u11_send_int(UCHARIS(S1D13U11_DISP_IF_0), S1D13U11_DEF_INT_MSG_LEN, lcdc_buf);
	if(ret){
		DPRINT("%s ERROR %d \n", __FUNCTION__, ret);
	}

	return ret;
}


//---------------------------------------------------------------------------- 
// s1d13u11_ph_probe()
// Probe for peripheral interface
// - Save usb interface data
// - Initialize peripheral driver
// 
// param[in]	intf	: struct of usb_interface
// param[in]	id		: struct of usb_device_id
//
// retval		0		: success
// retval  	not 0	: if error, then error code
// 
//---------------------------------------------------------------------------- 
static int s1d13u11_ph_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	int	ret = 0;
	
	DPRINT("%s \n", __FUNCTION__);

	lcdc[S1D13U11_DISP_IF_0] = kmalloc(sizeof(struct s1d13u11_vendor), GFP_KERNEL);
	if(lcdc[S1D13U11_DISP_IF_0]==(struct s1d13u11_vendor*)NULL){
		return -ENOMEM;
	}
	memset(lcdc[S1D13U11_DISP_IF_0], 0, sizeof(struct s1d13u11_vendor));

	lcdc[S1D13U11_DISP_IF_0]->intf=intf;
	lcdc[S1D13U11_DISP_IF_0]->devid=(struct usb_device_id*)id;
#ifndef PG_RELIEF
	lcdc[S1D13U11_DISP_IF_0]->usbdev=(struct usb_device*)interface_to_usbdev(intf);	/* pgr0431 */
#endif
	lcdc[S1D13U11_DISP_IF_0]->pipebout=(int)S1D13U11_DISP_EP_IF_0_BULK_OUT;
	lcdc[S1D13U11_DISP_IF_0]->pipebin=(int)S1D13U11_DISP_EP_IF_0_BULK_IN;
	lcdc[S1D13U11_DISP_IF_0]->pipeint=(int)S1D13U11_DISP_EP_IF_0_INT;
	lcdc[S1D13U11_DISP_IF_0]->pipedisp=(int)S1D13U11_DISP_EP_IF_0_DISP;

	/* initialize peripheral driver */
	ret=s1d13u11_init_peripheral();
	usb_set_intfdata(intf, lcdc[S1D13U11_DISP_IF_0]);

	return ret;
}


//---------------------------------------------------------------------------- 
// s1d13u11_ph_remove()
// Remove for peripheral interface
// - What usb device drivers talk to
// - USB device drivers dettach to interfaces on a physical device.
// 
// param[in]	intf	: struct of usb_interface
//
// retval		0		: success
// retval  	not 0	: if error, then error code
// 
//---------------------------------------------------------------------------- 
static int s1d13u11_ph_remove(struct usb_interface *intf)
{
	struct s1d13u11_vendor		*plcdc;
	struct s1d13u11_disp_event	eventStr;

	DPRINT("%s \n", __FUNCTION__);

	plcdc = usb_get_intfdata(intf);
	usb_set_intfdata(intf, NULL);
	if(plcdc){
		/* cancel INT transfer */
		if(plcdc->urb){
			usb_kill_urb(plcdc->urb);
			usb_free_urb(plcdc->urb);
		}

		/* free buffer */
		lcdc[S1D13U11_DISP_IF_0]=(struct s1d13u11_vendor*)NULL;
	}

	// Disconnect Notice
	memset( (void *)&eventStr, 0x00, sizeof(struct s1d13u11_disp_event) );

	eventStr.etype = S1D13U11_DISP_EVENT_INT_STAT_DISCONNECT ;

	if(lcdc_callback[S1D13U11_DEV_ID_INT][S1D13U11_EV_TYPE_INT] != (s1d13u11_callback_t)NULL) {
		lcdc_callback[S1D13U11_DEV_ID_INT][S1D13U11_EV_TYPE_INT](
						(unsigned int)sizeof(struct s1d13u11_disp_event),
						(unsigned char *)&eventStr );
	}

	return 0;
}


static struct work_struct work;

//---------------------------------------------------------------------------
// s1d13u11fb_init_work_routine()
//
//---------------------------------------------------------------------------
static void s1d13u11fb_init_work_routine(struct work_struct *work)
{
	msleep(1500);
	DPRINT("%s: start\n",__FUNCTION__);
	s1d13u11fb_init();
}


//---------------------------------------------------------------------------- 
// s1d13u11_probe()
// probe S1D13U11 Devices
// - Check input data.
// - "s1d13u11_ph_probe" call.
// - "Connect" Interrupt to Application.
// 
// param[in]	intf	: struct of usb_interface
// param[in]	id		: struct usb_device_id
//
// retval		0		: success
// retval  	not 0	: if error, then error code
// 
//---------------------------------------------------------------------------- 
static int s1d13u11_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	int	ret=0;
	struct s1d13u11_disp_event	eventStr;
	struct s1d13u11_disp_cbw_wakeup_config cbwData ;
	struct s1d13u11_disp_csw_wakeup_config cswData ;
	unsigned char *sendData ;
	int	allocSize = sizeof(struct s1d13u11_disp_cbw_wakeup_config) + sizeof(s_suspendData) ;

	DPRINT("%s starting probe \n", __FUNCTION__);

	if((intf==NULL)||(id==NULL)){
		return -EINVAL;
	}
	if(intf->altsetting==NULL){
		return -EINVAL;
	}
	if(&intf->altsetting->desc==NULL){
		printk("%s no descriptor \n", __FUNCTION__);
		return -EINVAL;
	}

	/* check interface number */
	DPRINT( "%s() checking interface number \n", __FUNCTION__ );
	switch(intf->altsetting->desc.bInterfaceNumber){
	case S1D13U11_DISP_IF_0:
		/* in case of peripheral */
		ret = s1d13u11_ph_probe(intf, id);

		// Connection Notice
		memset( (void *)&eventStr, 0x00, sizeof(struct s1d13u11_disp_event) );

		eventStr.etype = S1D13U11_DISP_EVENT_INT_STAT_CONNECT ;

		if(lcdc_callback[S1D13U11_DEV_ID_INT][S1D13U11_EV_TYPE_INT] != (s1d13u11_callback_t)NULL) {

			lcdc_callback[S1D13U11_DEV_ID_INT][S1D13U11_EV_TYPE_INT](
							(unsigned int)sizeof(struct s1d13u11_disp_event),
							(unsigned char *)&eventStr );
		}

		/* WAKEUP CONFIG Send */
		memset( (void *)&cbwData, 0x00, sizeof(struct s1d13u11_disp_cbw_wakeup_config) );
		memset( (void *)&cswData, 0x00, sizeof(struct s1d13u11_disp_csw_wakeup_config) );

		cbwData.cbw.cmd = S1D13U11_DISP_CMD_LCDC_WAKEUP_ON_CONFIG ;
		cbwData.bType = 0x00 ; // suspend
		cbwData.wWriteSize = sizeof(s_suspendData) ;

		sendData = (unsigned char *)kmalloc(allocSize, GFP_KERNEL );
		memset( (void *)sendData, 0x00, allocSize ) ;
		memcpy( sendData, &cbwData, sizeof(struct s1d13u11_disp_cbw_wakeup_config) );
		memcpy( (void *)(&sendData[16]), s_suspendData, sizeof(s_suspendData) );

		// Command Send
		ret = s1d13u11_control_cmd_send(
				allocSize,
				(unsigned char *)sendData,
				sizeof(struct s1d13u11_disp_csw_wakeup_config),
				(unsigned char *)&cswData );

		kfree( sendData );

		// status check
		if( cswData.csw.stat != S1D13U11_DISP_STAT_SUCCESS ) {
			// Status NG
			DPRINT( "ERROR-stat[%d] (%s-%d)\n",cswData.csw.stat,__FUNCTION__,__LINE__ );
			return -ENODEV ;
		}

		memset( (void *)&cbwData, 0x00, sizeof(struct s1d13u11_disp_cbw_wakeup_config) );
		memset( (void *)&cswData, 0x00, sizeof(struct s1d13u11_disp_csw_wakeup_config) );

		cbwData.cbw.cmd = S1D13U11_DISP_CMD_LCDC_WAKEUP_ON_CONFIG ;
		cbwData.bType = 0x01 ; // Resume
		cbwData.wWriteSize = sizeof(s_resumeData) ;

		allocSize = sizeof(struct s1d13u11_disp_cbw_wakeup_config) + sizeof(s_resumeData) ;

		sendData = (unsigned char *)kmalloc(allocSize, GFP_KERNEL );
		memset( (void *)sendData, 0x00, allocSize ) ;
		memcpy( sendData, &cbwData, sizeof(struct s1d13u11_disp_cbw_wakeup_config) );
		memcpy( (void *)(&sendData[16]), s_resumeData, sizeof(s_resumeData) );

		// Command Send
		ret = s1d13u11_control_cmd_send(
				allocSize,
				(unsigned char *)sendData,
				sizeof(struct s1d13u11_disp_csw_wakeup_config),
				(unsigned char *)&cswData );

		kfree( sendData );

		// status check
		if( cswData.csw.stat != S1D13U11_DISP_STAT_SUCCESS ) {
			// Status NG
			printk( "ERROR-stat[%d] (%s-%d)\n",cswData.csw.stat,__FUNCTION__,__LINE__ );
		}
		ret = 0 ;
		break;
	default:
		printk("%s unknown if %d \n", __FUNCTION__, intf->minor);
		ret = -ENODEV;
	}
	

	// let the user know what node this device is now attached to
	printk("S1D13U11 device now attached to USB-%d", intf->minor);
	
	//delay frame buffer initialization until after USB probe is complete
	INIT_WORK(&work, s1d13u11fb_init_work_routine);
	schedule_work(&work);
	
	DPRINT("%s: S1D13U11 sucessfully probed\n", __FUNCTION__);

	ret = 0 ;
	return ret;
}


//---------------------------------------------------------------------------- 
// s1d13u11_remove()
// remove S1D13U11
// - Check input data.
// - "s1d13u11_ph_remove" call.
// - "Disconnect" Interrupt to Application.
//
// param[in]	intf	: struct of usb_interface
//
// retval				: none
//
//---------------------------------------------------------------------------- 
static void s1d13u11_remove(struct usb_interface *intf)
{
	DPRINT("%s \n", __FUNCTION__);

	if(intf){
		/* check interface number */
		switch(intf->altsetting->desc.bInterfaceNumber){
		case S1D13U11_DISP_IF_0:
			/* in case of peripheral */
			s1d13u11_ph_remove(intf);
			DPRINT( "%s() removing peripherals \n", __FUNCTION__ );
			break;
		default:
			printk("%s unknown if %d \n", __FUNCTION__, intf->minor);
		}
	}
	return ;
}

#if 0
static int s1d13u11_wakeupcall(struct usb_interface *intf)
{
	printk( "(%s-%d)\n",__func__,__LINE__ );

	return 0;
}
#endif


// s1d13u11_id_table structure
static struct usb_device_id s1d13u11_id_table[] = {
	{USB_DEVICE(S1D13U11_DISP_ID_VENDOR, S1D13U11_DISP_ID_PRODUCT_DL)},
	{USB_DEVICE(S1D13U11_DISP_ID_VENDOR, S1D13U11_DISP_ID_PRODUCT_MK)},
	{}
};

MODULE_DEVICE_TABLE(usb, s1d13u11_id_table);


// s1d13u11_dev_driver structure
static struct usb_driver s1d13u11_driver = {
	.name		= S1D13U11_NAME_DRIVER,
	.probe		= s1d13u11_probe,
	.disconnect	= s1d13u11_remove,
	//.resume		= s1d13u11_wakeupcall,
	.id_table	= s1d13u11_id_table,
};


//---------------------------------------------------------------------------- 
// s1d13u11_init()
// Initialize S1D13U11 Driver
// - Initialize global.
// - Clear table.
// - Register driver
// 
// retval	0		: success
// retval  not 0	: if error, then error code
// 
//---------------------------------------------------------------------------- 
static int __init s1d13u11_init(void)
{
	int	ii ;
	int	jj ;
	int result;

	DPRINT("%s-%d \n", __FUNCTION__,__LINE__);

	spin_lock_init(&lcdc_lock);
	
#ifndef init_MUTEX
	sema_init(&lcdc_cmdlock, 1);
#else
	init_MUTEX( &lcdc_cmdlock );
#endif
#ifndef init_MUTEX
	sema_init(&lcdc_intlock, 1);
#else
	init_MUTEX( &lcdc_intlock );
#endif
	spin_lock_init(&lcdc_testint_lock);

	/* allocate buffer */
	lcdc_buf = kmalloc(S1D13U11_DEF_INT_BUF_LEN, GFP_KERNEL);
	DPRINT( "%s() lcdc_buf allocated \n", __FUNCTION__ );
	if(lcdc_buf){
		memset((char*)lcdc_buf, 0, S1D13U11_DEF_INT_BUF_LEN);

	}else{
		DPRINT("%s-%d \n", __FUNCTION__,__LINE__);
		return -ENOMEM;
	}

	/* initialize global */
	lcdc[S1D13U11_DISP_IF_0]=(struct s1d13u11_vendor*)NULL;
	lcdc[S1D13U11_DISP_IF_1]=(struct s1d13u11_vendor*)NULL;
	DPRINT( "%s() globals initialzed \n", __FUNCTION__ );

	init_waitqueue_head( (wait_queue_head_t *)&g_waitq );

	/* clear table */
	for(ii=0; ii<S1D13U11_DEV_ID_MAX; ii++){
		for(jj=0; jj<S1D13U11_EV_TYPE_MAX; jj++){
			lcdc_callback[ii][jj] = (s1d13u11_callback_t)NULL;
		}
	}

	/* register driver */
	DPRINT("%s-%d \n", __FUNCTION__,__LINE__);
	printk( "%s() registering driver \n", __FUNCTION__ );
	result = usb_register(&s1d13u11_driver);
	if (result)
		err("usb driver register failed. Error number %d", result);

	return result;
}

//---------------------------------------------------------------------------- 
// s1d13u11_exit()
// Exit S1D13U11 Driver
// - Free buffer.
// - Deregister driver
// 
// retval		:	none.
// 
//---------------------------------------------------------------------------- 
static void __exit s1d13u11_exit(void)
{
	DPRINT("(%s-%d)\n", __FUNCTION__,__LINE__);

	//---------------------------------------------------------------------
	/* free buffer */
	//---------------------------------------------------------------------
	if( lcdc_buf != (unsigned char*)NULL ) {
		kfree(lcdc_buf);
		lcdc_buf=(unsigned char*)NULL;
	}

	s1d13u11fb_exit();

	//---------------------------------------------------------------------
	/* deregister */
	//---------------------------------------------------------------------
	usb_deregister(&s1d13u11_driver);
	printk( "%s() deregistering driver \n", __FUNCTION__ );

	return ;
}

//---------------------------------------------------------------------------- 
// s1d13u11_control_cmd_send()
// Driver to Device Control Command Send
// - Send CBW message.
// - CBW message error check.
// - Send CSW message.
// - CSW message error check.
// 
// param[in]	send_len	: send length
// param[in]	send_data	: send data
// param[out]	recv_len	: receive length
// param[out]	recv_data	: receive data
//
// retval		0			: success
// retval		Not 0		: if error, then error code
//
//---------------------------------------------------------------------------- 
int s1d13u11_control_cmd_send(
			size_t		send_len,
			unsigned char *	send_data,
			size_t		recv_len,
			unsigned char *	recv_data )
{

//	DPRINT("Slen[%d] Rlen[%d](%s-%d)\n",send_len,recv_len,__FUNCTION__,__LINE__);

	int	ret = 0;

	S1D13U11_CMDLOCK;

	if( suspendFlg == 1 ){
		S1D13U11_CMDUNLOCK;
		return( -EFAULT );
	}

	// Input Parameter Check
	if( send_len < 1 ) {
		S1D13U11_CMDUNLOCK;
		return( -EFAULT );
	}
	if( (send_data==NULL) || (recv_data==NULL) ) {
		DPRINT("(%s-%d)\n", __FUNCTION__,__LINE__);
		S1D13U11_CMDUNLOCK;
		return( -EFAULT );
	}

	/* send CBW */
	ret = s1d13u11_send_cbw(
			UCHARIS(S1D13U11_DISP_IF_CMD),
			send_len,
			send_data,
			recv_len,
			recv_data );

	if( ret<0 ) {
		/* unlock */
		DPRINT("(%s-%d)\n", __FUNCTION__,__LINE__);
		S1D13U11_CMDUNLOCK;
		return( ret ) ;
	}

	/* receive CSW */
	ret = s1d13u11_recv_csw(
			UCHARIS(S1D13U11_DISP_IF_CMD),
			recv_len,
			recv_data );

	/* unlock */
	DPRINT("ret[%d](%s-%d)\n", ret,__FUNCTION__,__LINE__);
	S1D13U11_CMDUNLOCK;
	return ret ;
}


#ifdef CONFIG_S1D13U11_DEBUG_MESSAGES
//---------------------------------------------------------------------------- 
// dump_data()
// Debug Data USB-LCDC Peripheral core
// - Check input data.
// - Debug message output.
// 
// param[in]	msg		: message kind
// param[in]	data	: log data.
// param[in]	len		: data length.
//
// retval				: none.
//
//---------------------------------------------------------------------------- 
static void dump_data(char *msg, unsigned char *data, int len)
{
	int	i;

	if(data==(unsigned char*)NULL){
		return;
	}
	if(len>32){	//was 32
		len=32;
	}
	printk("%s \n",msg);
	for(i=0; i<len; i++){
		if((i!=0)&&((i%8)==0)){
			printk(" ");
		}
		if((i!=0)&&((i%16)==0)){
			printk("\n");
		}
		printk("%02X ", data[i]);
	}
	printk("\n");
}
#endif


module_init(s1d13u11_init);
module_exit(s1d13u11_exit);


EXPORT_SYMBOL(lcdc_lock);
EXPORT_SYMBOL(lcdc_cmdlock);
EXPORT_SYMBOL(lcdc_testint_lock);

EXPORT_SYMBOL(s1d13u11_cnv_size);
EXPORT_SYMBOL(s1d13u11_get_acctype);

EXPORT_SYMBOL(s1d13u11_send_cbw);
EXPORT_SYMBOL(s1d13u11_recv_csw);
EXPORT_SYMBOL(s1d13u11_send_data);
EXPORT_SYMBOL(s1d13u11_recv_data);
EXPORT_SYMBOL(s1d13u11_send_int);
EXPORT_SYMBOL(s1d13u11_recv_int);
EXPORT_SYMBOL(s1d13u11_clear_halt);
EXPORT_SYMBOL(s1d13u11_soft_reset);
EXPORT_SYMBOL(s1d13u11_suspend);
EXPORT_SYMBOL(s1d13u11_suspend_status);
EXPORT_SYMBOL(s1d13u11_busreset);
EXPORT_SYMBOL(s1d13u11_resume);
EXPORT_SYMBOL(s1d13u11_control_cmd_send);

MODULE_AUTHOR("USB display project team.");
MODULE_LICENSE("GPL v2");
MODULE_DESCRIPTION("SEIKO EPSON S1D13U11 LCDC Core Driver");
