//-----------------------------------------------------------------------------
// linux/drivers/usb/misc/epson-13u11/s1d13u11_fb.c  
//
// This file provides the setup routines for the Epson S1D13U11 frame buffer
// driver.
// 
// Copyright(C) SEIKO EPSON CORPORATION 2010-2013. All rights reserved.
//
// This driver software is distributed as is, without any warranty of any kind,
// either express or implied as further specified in the GNU Public License. This
// software may be used and distributed according to the terms of the GNU Public
// License, version 2. See the file COPYING in the main directory of this archive
// for more details.
//
// You should have received a copy of the GNU General Public License along with
// this program. If not, see <http://www.gnu.org/licenses/>.
//
//---------------------------------------------------------------------------- 
//
// WARNING: This file uses a chip specific include file. The include file 
// contains all hardware information needed to generate proper timings and 
// access registers. For timing and register details, see the S1D13U11
// Hardware Functional Specification.
//
//-----------------------------------------------------------------------------

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/smp.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/uaccess.h>
#include <linux/signal.h>
#include <linux/platform_device.h>
#include <linux/usb.h>
#include <linux/vmalloc.h>

#include <linux/proc_fs.h>

// Reads/writes are performed on a copy of the framebuffer in system memory.
// A timer periodically writes this copy to the "real" framebuffer in the S1D13U11.
// This copy is called a virtual framebuffer (see #define EPSON_VIRTUAL_FRAMEBUFFER)
#define EPSON_VIRTUAL_FRAMEBUFFER

#ifdef EPSON_VIRTUAL_FRAMEBUFFER
#include <linux/timer.h>
#endif

#include "s1d13u11_fb.h"
#include "s1d13u11.h"


#define S1D13U11FB_VERSION_STR		"EPSON13U11FB: $Revision: 3 $"

static struct semaphore		s_mutex ;


//-----------------------------------------------------------------------------
//
// Function Prototypes
//
//-----------------------------------------------------------------------------

#ifdef EPSON_VIRTUAL_FRAMEBUFFER
void s1d13u11fb_flush_virtual_framebuffer( struct fb_info *info );
static void s1d13u11fb_deferred_io(struct fb_info *info, struct list_head *pagelist);
#endif

static void s1d13u11fb_imageblit(struct fb_info *fbp, const struct fb_image *fbi);
static void s1d13u11fb_fillrect(struct fb_info *fbp, const struct fb_fillrect *fbr);
static void s1d13u11fb_copyarea(struct fb_info *fbp, const struct fb_copyarea *fbc);

//static ssize_t s1d13u11fb_put_display(struct fb_info *info, size_t count);
static int s1d13u11fb_write_display( unsigned short size, unsigned char *wdata );
static int s1d13u11fb_read_display( unsigned int size, unsigned long index, unsigned char *rdata );
//static int s1d13u11fb_init_display( void );	//add back in if not using 13U11P00C100
static int s1d13u11fb_enable_display( void );
static int s1d13u11fb_disable_display( void );
static int s1d13u11fb_enable_display_size( unsigned int rectSize );
static int s1d13u11fb_display_size_chg( unsigned int rectSize );

char *s1d13u11fb_version = S1D13U11FB_VERSION_STR;
static unsigned int			init_param = 0;
module_param(init_param, uint, S_IRUGO);
MODULE_PARM_DESC(init_param, "initialize LCDC registers, 0=no, 1=yes, default=no");

static struct fb_ops s1d13u11fb_ops =
{
	.owner			= THIS_MODULE,
	.fb_check_var	= s1d13u11fb_check_var,
	.fb_setcolreg	= s1d13u11fb_setcolreg,
	.fb_fillrect	= s1d13u11fb_fillrect,
	.fb_copyarea	= s1d13u11fb_copyarea,
	.fb_imageblit	= s1d13u11fb_imageblit,

	//.fb_mmap		= s1d13u11fb_mmap,		//uses default mmap routine

	.fb_ioctl		= s1d13u11fb_ioctl,
};


struct fb_fix_screeninfo s1d13u11fb_fix = 
{
	.id 			= "epson13u11fb", 
	.type			= FB_TYPE_PACKED_PIXELS,
	.visual			= FB_VISUAL_TRUECOLOR,
	.type_aux		= 0,
	.xpanstep		= 0,
	.ypanstep		= 0,
	.ywrapstep		= 0,
	.smem_len		= S1D13U11_FB_VFB_SIZE_BYTES,
	.line_length	= S1D13U11_FB_SIZE_X*S1D13U11_FB_DEF_VAR_BPP/8, 
	.accel			= FB_ACCEL_NONE,
};


#ifdef EPSON_VIRTUAL_FRAMEBUFFER

#ifndef CONFIG_S1D13U11_TIMER
#define CONFIG_S1D13U11_TIMER 20
#warning using default CONFIG_S1D13U11_TIMER
#endif
#define VIDEO_REFRESH_PERIOD   (HZ/CONFIG_S1D13U11_TIMER)

static struct fb_deferred_io s1d13u11fb_defio =
{
	.delay			= VIDEO_REFRESH_PERIOD,		//in HZ,
	.deferred_io	= s1d13u11fb_deferred_io,
};
#endif


//----------------------------------------------------------------------------
// s1d13u11fb_flush_virtual_framebuffer()
//
// PRIVATE FUNCTION:
// Flushes the whole virtual framebuffer to the S1D13U11 frame buffer.
//----------------------------------------------------------------------------
void s1d13u11fb_flush_virtual_framebuffer( struct fb_info *info )
{
	unsigned short	WriteWindowData[20] ;
	int 			ii = 0 ;
	int ret;

	DPRINT( "%s(): starting flush virtual buffer\n", __FUNCTION__ );

	// Rect Position Set
	memset( (void *)&WriteWindowData, 0x00, sizeof(WriteWindowData) );

	WriteWindowData[ii++] = 0x0052 ;	// REG[52h] Input Mode Register
	WriteWindowData[ii++] = 0x0000 ;	// Buffer1

	WriteWindowData[ii++] = 0x005A ;	// REG[5Ah] Write Window X Start Position Register
	WriteWindowData[ii++] = 0x0000 ;
	WriteWindowData[ii++] = 0x005C ;	// REG[5Ch] Write Window Y Start Position Register 0
	WriteWindowData[ii++] = 0x0000 ;
	WriteWindowData[ii++] = 0x005E ;	// REG[5Eh] Write Window Y Start Position Register 1
	WriteWindowData[ii++] = 0x0000 ;

	WriteWindowData[ii++] = 0x0060 ;	// REG[60h] Write Window X End Position Register
	WriteWindowData[ii++] = 0x004F ;
	WriteWindowData[ii++] = 0x0062 ;	// REG[62h] Write Window Y End Position Register 0
	WriteWindowData[ii++] = 0x003B ;
	WriteWindowData[ii++] = 0x0064 ;	// REG[64h] Write Window Y End Position Register 1
	WriteWindowData[ii++] = 0x0003 ;

	WriteWindowData[ii++] = 0x002A ;	// REG[2Ah] Display Mode Register
	WriteWindowData[ii++] = 0x0001 ;	// Buffer1
	WriteWindowData[ii++] = 0x0050 ;	// REG[50h] Display Control Register
	WriteWindowData[ii++] = 0x0080 ;	// Update internal registers

	// Set Rect Window Size to Register
	s1d13u11fb_write_display( (ii * sizeof(short)), (unsigned char *)&WriteWindowData ) ;

	// change RECT Size
	s1d13u11fb_display_size_chg( S1D13U11_FB_VFB_SIZE_BYTES );

	//--------------------------------------------------------------
	// Send Full Screen Data
	//--------------------------------------------------------------
	ret = s1d13u11_send_data(
				0x0,
				S1D13U11_FB_VFB_SIZE_BYTES,
				(unsigned char*) info->screen_base );

	if( ret ) {
		printk("%s ERROR(%d): Expected %x bytes, got %x bytes\n", __FUNCTION__,__LINE__, S1D13U11_FB_VFB_SIZE_BYTES, ret);
	}
	return;
}


//----------------------------------------------------------------------------
// s1d13u11fb_update_pages()
//
//----------------------------------------------------------------------------
static void s1d13u11fb_update_pages(struct fb_info *info, int offset, int numpages)
{
	unsigned char	*rectData;
	unsigned short	WriteWindowData[20] ;
	int 			ii = 0 ;
	int				ret;
	int				x1, x2, y1, y2;

	DPRINT( "%s(): starting update_pages\n", __FUNCTION__ );
	
	//setup rectangle to write based on offset and numpages
	x1 = 1;
	x2 = S1D13U11_FB_SIZE_X;
	y1 = ((offset/S1D13U11_FB_DEF_SIZE_BYTEPP)/S1D13U11_FB_SIZE_X) +1;
	y2 = (offset + (numpages*PAGE_SIZE))/(S1D13U11_FB_SIZE_X*S1D13U11_FB_DEF_SIZE_BYTEPP);

	if (y2>S1D13U11_FB_SIZE_Y)		//ensure PAGE_SIZE doesn't exceed display range
		y2=S1D13U11_FB_SIZE_Y;

	// Rect Position Set
	memset( (void *)&WriteWindowData, 0x00, sizeof(WriteWindowData) );

	WriteWindowData[ii++] = 0x0052 ;	// REG[52h] Input Mode Register
	WriteWindowData[ii++] = 0x0000 ;	// Buffer1

	WriteWindowData[ii++] = 0x005A ;	// REG[5Ah] Write Window X Start Position Register
	WriteWindowData[ii++] = ((x1/8) << 1) & 0xFE ;
	WriteWindowData[ii++] = 0x005C ;	// REG[5Ch] Write Window Y Start Position Register 0
	WriteWindowData[ii++] = (y1-1) >> 2 ;
	WriteWindowData[ii++] = 0x005E ;	// REG[5Eh] Write Window Y Start Position Register 1
	WriteWindowData[ii++] = (y1-1) & 0x0003 ;

	WriteWindowData[ii++] = 0x0060 ;	// REG[60h] Write Window X End Position Register
	WriteWindowData[ii++] = (((x2/8) - 1) << 1) & 0xFE ;
	WriteWindowData[ii++] = 0x0062 ;	// REG[62h] Write Window Y End Position Register 0
	WriteWindowData[ii++] = (y2-1) >> 2 ;
	WriteWindowData[ii++] = 0x0064 ;	// REG[64h] Write Window Y End Position Register 1
	WriteWindowData[ii++] = (y2-1) & 0x0003 ;

	WriteWindowData[ii++] = 0x002A ;	// REG[2Ah] Display Mode Register
	WriteWindowData[ii++] = 0x0001 ;	// Buffer1
	WriteWindowData[ii++] = 0x0050 ;	// REG[50h] Display Control Register
	WriteWindowData[ii++] = 0x0080 ;	// Update internal registers

	// Set Rect Window Size to Register
	s1d13u11fb_write_display( (ii * sizeof(short)), (unsigned char *)&WriteWindowData ) ;

	// change RECT Size
	s1d13u11fb_display_size_chg( (y2 - y1 + 1)*S1D13U11_FB_SIZE_X*S1D13U11_FB_DEF_SIZE_BYTEPP );

	//allocate buffer for rectangle data (in bytes)
	rectData = (unsigned char *)kmalloc((y2 - y1 + 2)*S1D13U11_FB_SIZE_X*S1D13U11_FB_DEF_SIZE_BYTEPP, GFP_KERNEL );
	memcpy( rectData, (unsigned char *) info->screen_base + ((y1-1) * S1D13U11_FB_SIZE_X*S1D13U11_FB_DEF_SIZE_BYTEPP), (y2 - y1 +1)*S1D13U11_FB_SIZE_X*S1D13U11_FB_DEF_SIZE_BYTEPP );

	//--------------------------------------------------------------
	// Send Full Screen Data
	//--------------------------------------------------------------
	ret = s1d13u11_send_data(
				0x0,
				(y2 - y1+1)*S1D13U11_FB_SIZE_X*S1D13U11_FB_DEF_SIZE_BYTEPP,
				rectData );

	if( ret ) {
		printk("%s ERROR(%d): Expected %x bytes, got %x bytes\n", __FUNCTION__,__LINE__, (y2 - y1+1)*S1D13U11_FB_SIZE_X*S1D13U11_FB_DEF_SIZE_BYTEPP, ret);
	}
	
	kfree(rectData);
	return;
}

//----------------------------------------------------------------------------
// s1d13u11fb_flush_dirty_rect()
//
//----------------------------------------------------------------------------
static void s1d13u11fb_flush_dirty_rect(struct fb_info *info)
{
	u32 y, linebytes;
	u32 stride = S1D13U11_FB_SIZE_X*S1D13U11_FB_DEF_SIZE_BYTEPP;
	int ret;
	struct s1d13u11fb_par *par = info->par;
	unsigned char	*rectData, *rectDataTmp;
	unsigned char	*pData;
	unsigned short	WriteWindowData[20] ;
	int 			ii = 0 ;
	unsigned short	alignedx1, alignedx2, alignedy1, alignedy2;

	DPRINT( "%s(): starting flush dirty rectangle\n", __FUNCTION__ );
	DPRINT(" %s: par->dx1=%x, par->dx2=%x, par->dy1=%x, par->dy2=%x\n", __FUNCTION__, par->dx1, par->dx2, par->dy1, par->dy2 );

	// Optimization: detect the whole screen update
	if (par->dx1 == 0 && par->dx2 == (S1D13U11_FB_SIZE_X) &&
	    par->dy1 == 0 && par->dy2 == (S1D13U11_FB_SIZE_Y))
	{
		s1d13u11fb_flush_virtual_framebuffer( info );
		return;
	}

	//check if par->dx1 is in 8 pixel resolution, else lower it
	//must be in sequence 1,9,17,...
	if (par->dx1==1)
		alignedx1 = par->dx1;
	else if ((par->dx1 >= 0) && (par->dx1 <= 8))
		alignedx1 = 1;
	else if ((par->dx1%8)==0)
		alignedx1 = ((par->dx1-8) & 0xFFFFFFF8)+1;
	else if ((par->dx1%8)==1)
		alignedx1 = par->dx1;
	else
		alignedx1 = ((par->dx1) & 0xFFFFFFF8)+1;

	//check if par->dx2 is in 8 pixel resolution, else raise it
	//must be in sequence 8,16,24,...
	if ((par->dx2%8)==0)
		alignedx2 = par->dx2;
	else
		alignedx2 = (par->dx2 + 0x8) & 0xFFFFFFF8;
	
	//don't need to meet any alignment restrictions
	alignedy1 = par->dy1;
	alignedy2 = par->dy2;

	// Rect Position Set
	memset( (void *)&WriteWindowData, 0x00, sizeof(WriteWindowData) );

	WriteWindowData[ii++] = 0x0052 ;	// REG[52h] Input Mode Register
	WriteWindowData[ii++] = 0x0000 ;	// Buffer1

	WriteWindowData[ii++] = 0x005A ;	// REG[5Ah] Write Window X Start Position Register
	WriteWindowData[ii++] = ((alignedx1/8) << 1) & 0xFE ;
	WriteWindowData[ii++] = 0x005C ;	// REG[5Ch] Write Window Y Start Position Register 0
	WriteWindowData[ii++] = (alignedy1-1) >> 2 ;
	WriteWindowData[ii++] = 0x005E ;	// REG[5Eh] Write Window Y Start Position Register 1
	WriteWindowData[ii++] = (alignedy1-1) & 0x0003 ;

	WriteWindowData[ii++] = 0x0060 ;	// REG[60h] Write Window X End Position Register
	WriteWindowData[ii++] = (((alignedx2 / 8) - 1) << 1) & 0xFE ;
	WriteWindowData[ii++] = 0x0062 ;	// REG[62h] Write Window Y End Position Register 0
	WriteWindowData[ii++] = (alignedy2-1) >> 2 ;
	WriteWindowData[ii++] = 0x0064 ;	// REG[64h] Write Window Y End Position Register 1
	WriteWindowData[ii++] = (alignedy2-1) & 0x0003 ;

	WriteWindowData[ii++] = 0x002A ;	// REG[2Ah] Display Mode Register
	WriteWindowData[ii++] = 0x0001 ;	// Buffer1
	WriteWindowData[ii++] = 0x0050 ;	// REG[50h] Display Control Register
	WriteWindowData[ii++] = 0x0080 ;	// Update internal registers

	// Set Rect Window Size to Register
	s1d13u11fb_write_display( (ii * sizeof(short)), (unsigned char *)&WriteWindowData ) ;

	linebytes = (alignedx2-alignedx1+1)*S1D13U11_FB_DEF_VAR_BPP/8;
	rectData = (unsigned char *)kmalloc(((alignedy2 - alignedy1)+2)*linebytes, GFP_KERNEL );
	if (rectData==NULL){
		printk("%s: ERROR: memory not allocated\n", __FUNCTION__);
		kfree(rectData);
		return;
	}

	//setup temp pointer into virtual frame buffer
	pData = (info->screen_base + ((alignedy1-1) * stride) + ((alignedx1-1)*2) );

	// change RECT Size
	s1d13u11fb_display_size_chg( (alignedy2 - alignedy1 + 1)*linebytes );

	//setup temp pointer for memcpy
	rectDataTmp = rectData;
	
	for (y = alignedy1; y <= alignedy2; y++)
	{
		memcpy( rectDataTmp, pData, linebytes );
		if (y < alignedy2){
			pData += stride;
			rectDataTmp += linebytes;
		}
	}
	
	// get rectangle data in rectData so set temp pointers to NULL
	pData = NULL;
	rectDataTmp = NULL;

	//--------------------------------------------------------------
	// Send Full Screen Data
	//--------------------------------------------------------------
	ret = s1d13u11_send_data(
				0x0,
				(alignedy2 - alignedy1 + 1)*linebytes,
				rectData );

	if( ret ) {
		printk("%s ERROR(%d): Expected %x bytes, got %x bytes\n", __FUNCTION__,__LINE__, (alignedy2 - alignedy1 + 1)*linebytes, ret);
	}


	kfree (rectData);
	return;
}


//----------------------------------------------------------------------------
// s1d13u11fb_deferred_io()
// Copy virtual framebuffer to display framebuffer.
// Called from the deferred io workqueue.
//
//----------------------------------------------------------------------------
static void s1d13u11fb_deferred_io(struct fb_info *info, struct list_head *pagelist)
{
	// It's possible to get here without anything on the pagelist via
	// s1d13u11_deferred_io_touch() or via a userspace fsync() invocation.
	
	if (!list_empty(pagelist))
	{
		int consecutive_pages = 0, dirty_pages = 0;
		int prev_index = -1;
		struct page *cur, *first;
		struct fb_deferred_io *fbdefio = info->fbdefio;
		
		// Walk the written page list and copy the data 
		list_for_each_entry( cur, &fbdefio->pagelist, lru )
		{
			dirty_pages++;
			
			if (prev_index == -1)
			{
				first = cur;
				consecutive_pages = 1;
			}
			else if (prev_index == cur->index - 1)
			{
				consecutive_pages++;
			}
			else
			{
				// Found a non-consecutive page. Update all consecutive pages first...
				s1d13u11fb_update_pages( info, (first->index << PAGE_SHIFT), consecutive_pages );
				first = cur;
				consecutive_pages = 1;
			}
			
			prev_index = cur->index;
		}
		
		// Last chunk of consecutive pages
		if (consecutive_pages)
			s1d13u11fb_update_pages( info, (first->index << PAGE_SHIFT), consecutive_pages );
		
		// If only a small number of pages were hit, adjust the delay period.
		// This assumes a graphic cursor being moved, so a cursor movement looks less jumpy.
		if (dirty_pages <= 10)
		{
			fbdefio->delay = 20;		//VIDEO_REFRESH_PERIOD/2;
		}
		else
		{
			fbdefio->delay = VIDEO_REFRESH_PERIOD;
		}
	}
	else
	{
		struct s1d13u11fb_par *par = info->par;
		
		if (par->dirtyrange)
		{
			// Process memory region created by s1d13u11_deferred_io_touch()
			s1d13u11fb_flush_dirty_rect( info );
			par->dirtyrange = 0;
		}
		else
		{
			printk( "s1d13u11fb_ideferred_io: NO PAGES\n" );
			s1d13u11fb_flush_virtual_framebuffer( info );
		}
	}
}

struct fb_rect {
	__u32 dx;	/* screen-relative */
	__u32 dy;
	__u32 width;
	__u32 height;
};


//----------------------------------------------------------------------------
// s1d13u11fb_deferred_io_touch()
//
//----------------------------------------------------------------------------
static void s1d13u11fb_deferred_io_touch(struct fb_info *info,const struct fb_rect* rect)
{
	struct fb_deferred_io *fbdefio = info->fbdefio;

	if (fbdefio)
	{
		struct s1d13u11fb_par *par = info->par;
		
		if (par->dirtyrange == 0)
		{
			par->dx1 = rect->dx;
			par->dy1 = rect->dy;
			par->dx2 = rect->dx+rect->width;
			par->dy2 = rect->dy+rect->height;
		}
		else
		{
			// Merge the new dirty range with the current one
			if (rect->dx < par->dx1)
				par->dx1 = rect->dx;
			if (rect->dy < par->dy1)
				par->dy1 = rect->dy;
			if (rect->dx+rect->width > par->dx2)
				par->dx2 = rect->dx+rect->width;
			if (rect->dy+rect->height > par->dy2)
				par->dy2 = rect->dy+rect->height;
		}

		DPRINT( "%s(): dr:%d dx1:%d dx2:%d dy1:%d dy2:%d\n", __FUNCTION__, par->dirtyrange,par->dx1,par->dx2,par->dy1,par->dy2 );

		par->dirtyrange = 1;
		schedule_delayed_work( &info->deferred_work, fbdefio->delay );
	}
}


//-----------------------------------------------------------------------------
// s1d13u11fb_probe()
// Initialize the chip and the frame buffer driver.
//
//-----------------------------------------------------------------------------
static int s1d13u11fb_probe(struct platform_device *pdev)
{
	struct fb_info *info;
	struct s1d13u11fb_par *par;

	DPRINT( "%s() called \n", __FUNCTION__ );

	info = framebuffer_alloc(sizeof(struct s1d13u11fb_par), &pdev->dev);
	if(!info)
	{
		return -ENOMEM;
	}
	
	par = info->par;
	
	// Initialize the chip
	// not required when using the S5U13U11P00C100 as it is already initialized
	//s1d13u11fb_init_display();

	// Allocate the virtual framebuffer
	//info->screen_base = (unsigned char *) kmalloc(S1D13U11_FB_VFB_SIZE_BYTES, GFP_KERNEL);
	info->screen_base = (unsigned char *) vmalloc(S1D13U11_FB_VFB_SIZE_BYTES);
	if (info->screen_base == NULL)
	{
		framebuffer_release(info);
		printk(KERN_WARNING "s1d13u11fb_probe: kmalloc() failed.\n");
		return -ENOMEM;
	}
	
	info->screen_size = S1D13U11_FB_VFB_SIZE_BYTES;
	info->fix = s1d13u11fb_fix;
	info->fix.mmio_start = (unsigned long) virt_to_phys((void*)par->RegAddr);
	info->fix.mmio_len = par->RegAddrMappedSize; 
	info->fix.smem_start = virt_to_phys((void*)info->screen_base); 
	info->fix.smem_len = S1D13U11_FB_VFB_SIZE_BYTES;
	
	DPRINT("%s: screen_base=%xn", __FUNCTION__, (unsigned int) info->screen_base );
	DPRINT("%s: screen_size=%x\n", __FUNCTION__, S1D13U11_FB_VFB_SIZE_BYTES );
	DPRINT("%s: fix.smem_start=%lx\n", __FUNCTION__, info->fix.smem_start );
	
	// Initialize Hardware Display Blank
	par->blank_mode = FB_BLANK_UNBLANK;
	//s1d13u11fb_ops.fb_blank = s1d13u11fb_blank;

	// Set flags for controller supported features
	info->flags = FBINFO_FLAG_DEFAULT|FBINFO_VIRTFB;

	// Set the pseudo palette
	info->pseudo_palette = par->pseudo_palette;

	s1d13u11fb_check_var( &info->var, info );
	info->fbops = &s1d13u11fb_ops;
	info->node = -1;

#ifdef EPSON_VIRTUAL_FRAMEBUFFER 
	info->fbdefio = &s1d13u11fb_defio;
	fb_deferred_io_init( info );
#endif

	if (register_framebuffer(info) < 0)
	{
		printk(KERN_WARNING "%s(): Could not register frame buffer with kernel.\n", __FUNCTION__ );
		framebuffer_release( info );
		kfree((void*) info->screen_base);
		return -EINVAL;
	}
	
#ifndef init_MUTEX
	sema_init(&s_mutex, 1);
#else
	init_MUTEX( &s_mutex );
#endif

	printk( "fb%d: %s frame buffer device\n", info->node, info->fix.id );

	return 0;
}


//-----------------------------------------------------------------------------
// Use system Blit routines.
//-----------------------------------------------------------------------------

void s1d13u11fb_fillrect(struct fb_info *info, const struct fb_fillrect *rect)
{
	sys_fillrect(info,rect);
#ifdef VIDEO_REFRESH_PERIOD 
	s1d13u11fb_deferred_io_touch( info, (const struct fb_rect *) rect );
#endif
}

void s1d13u11fb_copyarea(struct fb_info *info, const struct fb_copyarea *area)
{
	sys_copyarea(info,area);
#ifdef VIDEO_REFRESH_PERIOD 
	s1d13u11fb_deferred_io_touch( info, (const struct fb_rect *) area);
#endif
}

void s1d13u11fb_imageblit(struct fb_info *info, const struct fb_image *image)
{
	sys_imageblit(info, image);
#ifdef VIDEO_REFRESH_PERIOD 
	s1d13u11fb_deferred_io_touch( info, (const struct fb_rect *) image );
#endif
}


//-----------------------------------------------------------------------------
// s1d13u11fb_remove()
// Safely shutdown the frame buffer driver.
//
//-----------------------------------------------------------------------------
//#ifdef CONFIG_S1D13U11_USB
static int s1d13u11fb_remove(struct platform_device *dev)
//#else
//static int __devexit s1d13u11fb_remove(struct platform_device *dev)
//#endif
{
	struct fb_info *info = platform_get_drvdata( dev );

	if (info)
	{
		s1d13u11fb_CleanUp();

		unregister_framebuffer( info );

	#ifdef VIDEO_REFRESH_PERIOD
		fb_deferred_io_cleanup( info );
	#endif

	#ifdef EPSON_VIRTUAL_FRAMEBUFFER
		kfree( (void *) info->screen_base );
	#endif

		framebuffer_release( info );
	}

	return 0;
}


static struct platform_driver s1d13u11fb_driver = {
	.probe	= s1d13u11fb_probe,
	.remove = s1d13u11fb_remove,
	.driver	= {
		.owner	= THIS_MODULE,
		.name	= "s1d13u11fb",
	},
};


static struct platform_device *s1d13u11fb_device;


//-----------------------------------------------------------------------------
// s1d13u11fb_init()
//
//-----------------------------------------------------------------------------
int s1d13u11fb_init(void)
{
	int ret;
	
	printk( "%s():", __FUNCTION__ );
	ret = platform_driver_register( &s1d13u11fb_driver );

	if (!ret)
	{
		s1d13u11fb_device = platform_device_register_simple( "s1d13u11fb", 0, NULL, 0 );
		if (IS_ERR(s1d13u11fb_device))
		{
			printk( "%s(): error\n", __FUNCTION__ );
			platform_driver_unregister( &s1d13u11fb_driver );
			ret = PTR_ERR( s1d13u11fb_device );
		}
	}

	return ret;
}


//-----------------------------------------------------------------------------
// s1d13u11fb_exit()
//
//-----------------------------------------------------------------------------
void __exit s1d13u11fb_exit(void)
{
	platform_device_unregister( s1d13u11fb_device );
	platform_driver_unregister( &s1d13u11fb_driver );
}


//----------------------------------------------------------------------------
// s1d13u11fb_CleanUp()
//
//----------------------------------------------------------------------------
void s1d13u11fb_CleanUp(void)
{
}


//----------------------------------------------------------------------------- 
// s1d13u11fb_check_var()
// Fill in the 'var' and 'fix' structures. 
//
//----------------------------------------------------------------------------- 
int s1d13u11fb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
	var->xres			= S1D13U11_FB_SIZE_X;
	var->yres			= S1D13U11_FB_SIZE_Y;
	var->xres_virtual	= var->xres;
	var->yres_virtual	= var->yres;
	var->xoffset		= var->yoffset = 0;
	var->bits_per_pixel	= S1D13U11_FB_DEF_VAR_BPP;
#ifndef MFB
	var->grayscale		= 0;
#else
	var->grayscale		= 1;
#endif
	var->nonstd 		= 0;					// != 0 Non standard pixel format
	var->activate		= FB_ACTIVATE_NOW;		//see FB_ACTIVATE_*
	var->height 		= -1;					// height of picture in mm
	var->width			= -1;					// width of picture in mm
	var->accel_flags	= 0;					// acceleration flags (hints
	//var->pixclock		= S1D_DISPLAY_PCLK;
	var->right_margin	= 0;
	var->lower_margin	= 0;
	var->hsync_len		= 0;
	var->vsync_len		= 0;
	var->left_margin	= 0;
	var->upper_margin	= 0;
	var->sync			= 0;
	var->vmode			= FB_VMODE_NONINTERLACED;
	var->red.msb_right 	= var->green.msb_right = var->blue.msb_right = 0;
	var->transp.offset 	= var->transp.length = var->transp.msb_right = 0;

	s1d13u11fb_fix.visual	= FB_VISUAL_TRUECOLOR;
  
	switch (info->var.bits_per_pixel)
	{
	  case 8:
		var->red.offset  = var->green.offset = var->blue.offset = 0;
		var->red.length  = var->green.length = var->blue.length = S1D13U11_FB_DEF_VAR_BPP;
		break;

	  case 16:
		var->red.offset 	= 11;
		var->red.length 	= 5;
		var->green.offset	= 5;
		var->green.length	= 6;
		var->blue.offset	= 0;
		var->blue.length	= 5;
		var->transp.offset	= 0;
		var->transp.length	= 0;
		break;
		
	  case 24:	//24bpp
		var->red.offset 	= 16;
		var->red.length 	= 8;
		var->green.offset	= 8;
		var->green.length	= 8;
		var->blue.offset	= 0;
		var->blue.length	= 8;
		var->transp.offset	= 0;
		var->transp.length	= 0;
		break;

	  default:
		printk( KERN_WARNING "%dbpp is not supported.\n", info->var.bits_per_pixel );
		return -EINVAL;
	}

	return 0;
}


//---------------------------------------------------------------------------- 
// s1d13u11fb_setcolreg()
//
// Set a single color register. The values supplied have a 16 bit magnitude.  
// Return != 0 for invalid regno.  
//
// We get called even if we specified that we don't have a programmable palette
// or in direct/true color modes!
//
//---------------------------------------------------------------------------- 
int s1d13u11fb_setcolreg(unsigned regno, unsigned red, unsigned green, unsigned blue, unsigned transp, struct fb_info *info) 
{
	DPRINT( "%s(): regno=%02Xh red=%04Xh green=%04Xh blue=%04Xh transp=%04Xh\n", __FUNCTION__, regno, red, green, blue, transp );

	if (regno >= S1D13U11_FB_PALETTE_SIZE) 
	{
		printk(KERN_WARNING "%s(): reg index too large (%Xh)\n", __FUNCTION__, regno);
		return 1; 
	}

	// setup pseudo palette values
	// Make the first 16 LUT entries available to the console
	switch (info->var.bits_per_pixel)
	{
	  case 8:
		if ( s1d13u11fb_fix.visual == FB_VISUAL_TRUECOLOR )
		{
			if (regno < 16)
			{	/* RGB 332 */
				((u32*)info->pseudo_palette)[regno] = ((red & 0xe000)>>8)|((green & 0xe000) >> 11)|((blue & 0xc000) >> 14); 
			}
		}
		break;

	  default:
	  case 16:
		if (regno < 16)
		{	/* RGB 565 */
			((u32*)info->pseudo_palette)[regno] = (red & 0xf800)|((green & 0xfc00) >> 5)|((blue & 0xf800) >> 11); 
		}
		break;
		
	  case 24:		//24bpp
		if (regno < 16)
		{	/* RGB 888 */
			((u32*)info->pseudo_palette)[regno] = (red & 0xff00)|((green & 0xff00) >> 8)|((blue & 0xff00) >> 16); 
		}
		break;
	}

	return 0; 
}


//---------------------------------------------------------------------------- 
// s1d13u11fb_ioctl()
// Handles IOCTL messages from the user space.
//
//---------------------------------------------------------------------------- 
int s1d13u11fb_ioctl(struct fb_info *info, unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	int						ret = 0;
	unsigned char			*cmdCbwDat;
	unsigned char			*cmdCswDat;
	struct s1d13u11_ctl		*pctl;
	
	S1D13U11LCDC_LOCK ;

	DPRINT("%s(): cmd=%04Xh\n", __FUNCTION__, cmd);

	down_interruptible( &s_mutex );
	
	// frame buffer info
	if(info==(struct fb_info*)NULL){
		printk("%s: no fb_info \n", __FUNCTION__);
		up( &s_mutex );
		S1D13U11LCDC_UNLOCK ;
		return -EINVAL;
	}

	switch(cmd) {
	// manually flushes the contents of the virtual frame buffer
	case S1D13U11_FB_CMD_FLUSH_VFB:
		s1d13u11fb_flush_virtual_framebuffer(info);
		break;
		
	case S1D13U11_FB_CMD_LCDC_WRITE:
		pctl = (struct s1d13u11_ctl*)arg;

		cmdCbwDat = (unsigned char *)kmalloc( pctl->size, GFP_KERNEL );
		if( cmdCbwDat == (unsigned char *)NULL ) {
			return -ENOMEM ;
		}

		/* copy data from user */
		ret = (int)copy_from_user( cmdCbwDat, pctl->buf, pctl->size);
		if( ret ) {
			kfree( cmdCbwDat );
			return -EFAULT;
		}

		ret = s1d13u11fb_write_display( pctl->size, cmdCbwDat );

		kfree( cmdCbwDat );
		break ;

	//allows a single register to be read
	case S1D13U11_FB_CMD_LCDC_READ:
		pctl = (struct s1d13u11_ctl*)arg;

		cmdCswDat = (unsigned char *)kmalloc( pctl->size, GFP_KERNEL );
		if( cmdCswDat == (unsigned char *)NULL ) {
			return -ENOMEM ;
		}

		ret = s1d13u11fb_read_display( pctl->size, pctl->addr, cmdCswDat );
		pctl->value = cmdCswDat[0];

		copy_to_user(argp, &pctl, sizeof(pctl)) ? -EFAULT : 0;
		break;

	default:
		printk("%s: unknown cmd %d \n", __FUNCTION__, cmd);
		up( &s_mutex );

		S1D13U11LCDC_UNLOCK ;
		return -EINVAL;
	}

	up( &s_mutex );

	S1D13U11LCDC_UNLOCK ;
	
	return ret;
}

#if 0
//---------------------------------------------------------------------------- 
// s1d13u11fb_put_display()
// Framebuffer Driver image to display.
//
//---------------------------------------------------------------------------- 
static ssize_t s1d13u11fb_put_display( struct fb_info *info, size_t count )
{
	int	ret = 0;

	//--------------------------------------------------------------
	// Send Full Screen Data
	//--------------------------------------------------------------
	ret = s1d13u11_send_data(
				S1D13U11_DISP_IF_0,
				S1D13U11_FB_SIZE_WR_DATA,
				(unsigned char*) info->screen_base );

	if( ret!=(short)S1D13U11_FB_SIZE_WR_DATA ) {
		printk("%s ERROR put display(%d) \n", __FUNCTION__,__LINE__);
	}

	return ret;
}
#endif


//---------------------------------------------------------------------------- 
// s1d13u11fb_init_display()
// Framebuffer Driver Device Initialize.
//
//---------------------------------------------------------------------------- 
static int s1d13u11fb_init_display( void )
{
	int	ret = 0;

	DPRINT( "(%s-%d)\n",__FUNCTION__,__LINE__ );

	if (init_param) {
		s1d13u11fb_write_display(
				(unsigned short)sizeof(s_initData),
				(unsigned char *)&s_initData );
	} else {
		s1d13u11fb_write_display(
				(unsigned short)sizeof(s_setData),
				(unsigned char *)&s_setData );
	}
	s1d13u11fb_write_display(
			(unsigned short)sizeof(s_initLCDData),
			(unsigned char *)&s_initLCDData );

	return ret;
}


//---------------------------------------------------------------------------- 
// s1d13u11fb_write_display()
// Framebuffer Driver Device Write.
//
//---------------------------------------------------------------------------- 
static int s1d13u11fb_write_display( unsigned short size, unsigned char *wdata )
{
	int	ret = 0 ;
	struct s1d13u11_disp_cbw_lcdc_write	cbwWData ;
	struct s1d13u11_disp_csw_lcdc_write	cswWData ;
	unsigned char *sendData ;

	DPRINT( "SIZE[%d] (%s-%d)\n",size,__FUNCTION__,__LINE__ );

	//---------------------------------------------------------------------
	// LCDC_VRAM_ACC_DISABLE
	//---------------------------------------------------------------------
	ret = s1d13u11fb_disable_display();
	if( ret!=0 ) {
		printk( "ERROR[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );
		return( ret );
	}

	//---------------------------------------------------------------------
	// DEVICE WRITE!!
	//---------------------------------------------------------------------
	memset( (void *)&cbwWData, 0x00, sizeof(cbwWData) );
	memset( (void *)&cswWData, 0x00, sizeof(cswWData) );

	cbwWData.cbw.cmd = S1D13U11_DISP_CMD_LCDC_WRITE ;
	cbwWData.size = size ;

	DPRINT( "wdata[%X %X %X %X %X %X %X %X](%s-%d)\n",wdata[0],wdata[1],wdata[2],wdata[3],wdata[4],
		wdata[5],wdata[6],wdata[7],__FUNCTION__,__LINE__ );
	DPRINT( "SIZE[%d] [%d] (%s-%d)\n",sizeof(struct s1d13u11_disp_cbw_lcdc_write),size,__FUNCTION__,__LINE__ );

	sendData = (unsigned char *)kmalloc((sizeof(struct s1d13u11_disp_cbw_lcdc_write) + size), GFP_KERNEL );

	if( sendData<0 ) {
		// Send NG
		printk( "ENOMEM(%s-%d)\n",__func__,__LINE__ );
		return( -ENOMEM );
	}

	memset( (void *)sendData, 0x00, (sizeof(struct s1d13u11_disp_cbw_lcdc_write) + size) ) ;
	memcpy( sendData, &cbwWData, sizeof(struct s1d13u11_disp_cbw_lcdc_write) );
	memcpy( (void *)(&sendData[16]), wdata, size );

	// Command Send
	ret = s1d13u11_control_cmd_send(
			(sizeof(struct s1d13u11_disp_cbw_lcdc_write) + size),
			(unsigned char *)sendData,
			sizeof(struct s1d13u11_disp_csw_lcdc_write),
			(unsigned char *)&cswWData );

	kfree( sendData );

	if( ret<0 ) {
		// Send NG
		s1d13u11fb_enable_display();
		printk( "ERROR[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );
		return( -EFAULT );
	}

	// status check
	if( cswWData.csw.stat != S1D13U11_DISP_STAT_SUCCESS ) {
		// Status NG
		DPRINT( "SIZE[%d] [%d] (%s-%d)\n",sizeof(struct s1d13u11_disp_cbw_lcdc_write),size,__FUNCTION__,__LINE__ );
		printk( "ERROR-stat[%d] (%s-%d)\n",cswWData.csw.stat,__FUNCTION__,__LINE__ );
		return( -EFAULT );
	}

	//---------------------------------------------------------------------
	// LCDC_VRAM_ACC_ENABLE
	//---------------------------------------------------------------------
	ret = s1d13u11fb_enable_display();

	DPRINT( "RETURN[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );

	return( ret ) ;
}


//---------------------------------------------------------------------------- 
// s1d13u11fb_read_display()
// Framebuffer Driver Device Read.
//
//---------------------------------------------------------------------------- 
static int s1d13u11fb_read_display( unsigned int size, unsigned long index, unsigned char *rdata )
{
	int	ret = 0 ;
	int i;
	struct s1d13u11_disp_cbw_lcdc_read	cbwRData ;
	struct s1d13u11_disp_csw_lcdc_read	cswRData ;
	unsigned char *sendData ;
	unsigned char *recvData ;

	DPRINT( "SIZE[%d] (%s-%d)\n",size,__FUNCTION__,__LINE__ );

	//---------------------------------------------------------------------
	// LCDC_VRAM_ACC_DISABLE
	//---------------------------------------------------------------------
	ret = s1d13u11fb_disable_display();
	if( ret!=0 ) {
		printk( "ERROR[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );
		return( ret );
	}

	//---------------------------------------------------------------------
	// DEVICE READ
	//---------------------------------------------------------------------
	memset( (void *)&cbwRData, 0x00, sizeof(cbwRData) );
	memset( (void *)&cswRData, 0x00, sizeof(cswRData) + cbwRData.rSize);

	cbwRData.cbw.cmd = S1D13U11_DISP_CMD_LCDC_READ ;
	cbwRData.disInc = 0x00 ;						//signals register access
	cbwRData.rSize = (unsigned short) size ;		//number of registers to read
	cbwRData.rAdd = (unsigned short) index ;		//index to start reading from

	DPRINT("disInc=%x, rSize=%x, rAdd=%x\n", cbwRData.disInc, cbwRData.rSize, cbwRData.rAdd);

	sendData = (unsigned char *)kmalloc((int)(sizeof(struct s1d13u11_disp_cbw_lcdc_read)), GFP_KERNEL );

	if( sendData<0 ) {
		// Send NG
		printk( "ENOMEM(%s-%d)\n",__FUNCTION__,__LINE__ );
		return( -ENOMEM );
	}

	memset( (void *)sendData, 0x00, (sizeof(struct s1d13u11_disp_cbw_lcdc_read)) ) ;
	memcpy( sendData, &cbwRData, sizeof(struct s1d13u11_disp_cbw_lcdc_read) );

	recvData = (unsigned char *)kmalloc((sizeof(struct s1d13u11_disp_csw_lcdc_read) + cbwRData.rSize), GFP_KERNEL );

	// Command Send
	ret = s1d13u11_control_cmd_send(
			sizeof(struct s1d13u11_disp_cbw_lcdc_read),
			(unsigned char *)sendData,
			(sizeof(struct s1d13u11_disp_csw_lcdc_read) + cbwRData.rSize),
			(unsigned char *)recvData );

	kfree( sendData );
	
	memcpy( &cswRData, recvData, sizeof(struct s1d13u11_disp_csw_lcdc_read) );
	
	if( ret<0 ) {
		// Send NG
		printk( "ERROR[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );
		return( -EFAULT );
	}

	// status check
	if( cswRData.csw.stat != S1D13U11_DISP_STAT_SUCCESS ) {
		// Status NG
		printk( "ERROR-stat[%d] (%s-%d)\n",cswRData.csw.stat,__FUNCTION__,__LINE__ );
		return( -EFAULT );
	}
	
	for (i=0; i<cbwRData.rSize; i++)
	{
		rdata[i] = recvData[sizeof(struct s1d13u11_disp_csw_lcdc_read)+i];
		DPRINT("%s: Read Register Data %x = %x, recvData=%x\n", __FUNCTION__, i, rdata[i], recvData[sizeof(struct s1d13u11_disp_csw_lcdc_read)+i]);
	}

	kfree( recvData );

	//---------------------------------------------------------------------
	// LCDC_VRAM_ACC_ENABLE
	//---------------------------------------------------------------------
	ret = s1d13u11fb_enable_display();

	DPRINT( "RETURN[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );

	return( ret ) ;
}


//---------------------------------------------------------------------------- 
// s1d13u11fb_disable_display()
// Framebuffer Driver VRAM Access disable
//
//---------------------------------------------------------------------------- 
static int s1d13u11fb_disable_display( void )
{
	int	ret = 0 ;
	struct s1d13u11_disp_cbw_lcdc_vram_acc_disable	cbwDisData ;
	struct s1d13u11_disp_csw_lcdc_vram_acc_disable	cswDisData ;

	DPRINT("%s START\n", __FUNCTION__);

	//---------------------------------------------------------------------
	// LCDC_VRAM_ACC_DISABLE
	//---------------------------------------------------------------------
	memset( (void *)&cbwDisData, 0x00, sizeof(cbwDisData) );
	memset( (void *)&cswDisData, 0x00, sizeof(cswDisData) );

	cbwDisData.cbw.cmd = S1D13U11_DISP_CMD_LCDC_VRAM_ACC_DISABLE ;

	// Command Send
	ret = s1d13u11_control_cmd_send(
			sizeof(struct s1d13u11_disp_cbw_lcdc_vram_acc_disable),
			(unsigned char *)&cbwDisData,
			sizeof(struct s1d13u11_disp_csw_lcdc_vram_acc_disable),
			(unsigned char *)&cswDisData );

	if( ret<0 ) {
		// Send NG
		printk( "ERROR[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );
		return( -EFAULT );
	}

	// status check
	if( cswDisData.csw.stat != S1D13U11_DISP_STAT_SUCCESS ) {
		// Status NG
		printk( "ERROR[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );
		return( -EFAULT );
	}

	return( 0 ) ;
}


//---------------------------------------------------------------------------- 
// s1d13u11fb_enable_display()
// Framebuffer Driver VRAM Access enable
//
//---------------------------------------------------------------------------- 
static int s1d13u11fb_enable_display( void )
{
	int	ret = 0 ;
	struct s1d13u11_disp_cbw_lcdc_vram_acc_enable	cbwEnaData ;
	struct s1d13u11_disp_csw_lcdc_vram_acc_enable	cswEnaData ;

	DPRINT("%s START\n", __FUNCTION__);

	//---------------------------------------------------------------------
	// LCDC_VRAM_ACC_ENABLE
	//---------------------------------------------------------------------
	memset( (void *)&cbwEnaData, 0x00, sizeof(cbwEnaData) );
	memset( (void *)&cswEnaData, 0x00, sizeof(cswEnaData) );

	cbwEnaData.cbw.cmd     = S1D13U11_DISP_CMD_LCDC_VRAM_ACC_ENABLE ;
	cbwEnaData.pictureSize = S1D13U11_FB_SIZE_WR_DATA ;

	// Command Send
	ret = s1d13u11_control_cmd_send(
			sizeof(struct s1d13u11_disp_cbw_lcdc_vram_acc_enable),
			(unsigned char *)&cbwEnaData,
			sizeof(struct s1d13u11_disp_csw_lcdc_vram_acc_enable),
			(unsigned char *)&cswEnaData );

	if( ret<0 ) {
		// Send NG
		printk( "ERROR[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );
		return( -EFAULT );
	}

	// status check
	if( cswEnaData.csw.stat != S1D13U11_DISP_STAT_SUCCESS ) {
		// Status NG
		printk( "ERROR[%d][%d] (%s-%d)\n",ret,cswEnaData.csw.stat,__FUNCTION__,__LINE__ );
		return( -EFAULT );
	}

	return( 0 ) ;
}


//---------------------------------------------------------------------------- 
// s1d13u11fb_enable_display_size()
// Framebuffer Driver VRAM Access enable (Size change)
//
//---------------------------------------------------------------------------- 
static int s1d13u11fb_enable_display_size( unsigned int rectSize )
{
	int	ret = 0 ;
	struct s1d13u11_disp_cbw_lcdc_vram_acc_enable	cbwEnaData ;
	struct s1d13u11_disp_csw_lcdc_vram_acc_enable	cswEnaData ;

	DPRINT("%s START\n", __FUNCTION__);

	//---------------------------------------------------------------------
	// LCDC_VRAM_ACC_ENABLE
	//---------------------------------------------------------------------
	memset( (void *)&cbwEnaData, 0x00, sizeof(cbwEnaData) );
	memset( (void *)&cswEnaData, 0x00, sizeof(cswEnaData) );

	cbwEnaData.cbw.cmd     = S1D13U11_DISP_CMD_LCDC_VRAM_ACC_ENABLE ;
	cbwEnaData.pictureSize = rectSize ;

	// Command Send
	ret = s1d13u11_control_cmd_send(
			sizeof(struct s1d13u11_disp_cbw_lcdc_vram_acc_enable),
			(unsigned char *)&cbwEnaData,
			sizeof(struct s1d13u11_disp_csw_lcdc_vram_acc_enable),
			(unsigned char *)&cswEnaData );

	if( ret<0 ) {
		// Send NG
		DPRINT( "ERROR[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );
		return( -EFAULT );
	}

	// status check
	if( cswEnaData.csw.stat != S1D13U11_DISP_STAT_SUCCESS ) {
		// Status NG
		DPRINT( "ERROR[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );
		return( -EFAULT );
	}

	return( 0 ) ;
}


//---------------------------------------------------------------------------- 
// s1d13u11fb_display_size_chg()
// Framebuffer Driver Device Write.
//
//---------------------------------------------------------------------------- 
static int s1d13u11fb_display_size_chg( unsigned int rectSize )
{
	int	ret = 0 ;

	DPRINT("%s START\n", __FUNCTION__);
	DPRINT( "SIZE[%d] (%s-%d)\n",rectSize,__FUNCTION__,__LINE__ );

	//---------------------------------------------------------------------
	// LCDC_VRAM_ACC_DISABLE
	//---------------------------------------------------------------------
	ret = s1d13u11fb_disable_display();
	if( ret!=0 ) {
		DPRINT( "ERROR[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );
		return( ret );
	}

	//---------------------------------------------------------------------
	// LCDC_VRAM_ACC_ENABLE
	//---------------------------------------------------------------------
	ret = s1d13u11fb_enable_display_size( rectSize );

	DPRINT( "RETURN[%d] (%s-%d)\n",ret,__FUNCTION__,__LINE__ );

	return( ret ) ;
}


EXPORT_SYMBOL(s1d13u11fb_init);
EXPORT_SYMBOL(s1d13u11fb_exit);

MODULE_AUTHOR("SEIKO EPSON CORPORATION");
MODULE_DESCRIPTION("EPSON S1D13U11 frame buffer driver");
MODULE_LICENSE("GPL v2");

