//-----------------------------------------------------------------------------
// linux/drivers/usb/misc/epson-13u11/s1d13u11_ts.c  
//
// This file provides the routines for the Touchscreen on the S5U13U11P00C
// evauluation board.
// 
// Copyright(C) SEIKO EPSON CORPORATION 2009-2013. All rights reserved.
//
// This driver software is distributed as is, without any warranty of any kind,
// either express or implied as further specified in the GNU Public License. This
// software may be used and distributed according to the terms of the GNU Public
// License, version 2. See the file COPYING in the main directory of this archive
// for more details.
//
// You should have received a copy of the GNU General Public License along with
// this program. If not, see <http://www.gnu.org/licenses/>.
//
//---------------------------------------------------------------------------- 

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/input.h>
#include <linux/timer.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/bitops.h>

#include "s1d13u11.h"


struct tp_event {
	struct s1d13u11_disp_event_spiseq_event	header ;
	/* For portability, we can't read 12 bit values using SPI (which
	 * would make the controller deliver them as native byteorder u16
	 * with msbs zeroed).  Instead, we read them as two 8-bit values,
	 * which need byteswapping then range adjustment.
	 */
	__be16 y;
	__be16 z1;
	__be16 z2;
	__be16 x;
};


#define					MAXTPEVENT	3
static struct tp_event	tpBuff[MAXTPEVENT] ;
static int				s_tpBuffCnt = 0 ;


// Touch Panel configuration defines
#define	TP_X_MIN			0		// X MIN POSITION
#define	TP_X_MAX			4095	// X MAX POSITION
#define	TP_Y_MIN			0		// Y MIN POSITION
#define	TP_Y_MAX			4095	// Y MAX POSITION
#define	TP_Z_MIN			0		// Z MIN POSITION
#define	TP_Z_MAX			4095	// Z MAX POSITION

#define	TP_SUSPEND_TIM		1000 	// msec(10sec)
#define	TP_CYCLE_TIM		3		// msec

#define	TP_SEQUENCE_DATACNT	8	// Sequence start data count
#define	S1D13U11_BL_NAME_DRIVER	"S1D13U11 Touchscreen"

#define TP_SEQUENCE_DELAY	0x00
#define TP_SEQUENCE_IRQ		0x01
#define	TP_SEQUENCE_NONE	0x02

#define	TP_SEQUENCE_WRITE	0x00
#define	TP_SEQUENCE_READ	0x01
#define	TP_SEQUENCE_CSENB	0x02
#define	TP_SEQUENCE_CSDIS	0x03
#define	TP_SEQUENCE_INTDIS	0x04
#define	TP_SEQUENCE_INTDNA	0x05
#define	TP_SEQUENCE_WAIT	0x06

static struct input_dev	*s_indev;

// prototype
unsigned int s1d13u11_tp_touch_read( unsigned short *len, unsigned char *buf );
unsigned int s1d13u11_tp_seq_read( unsigned short *len,	unsigned char *buf );
static int s1d13u11_tp_seq_stop( unsigned char spich );

// Touch Panel Module Parameter
static char* initpos = "initpos" ;
static char  gPendown = 0x00 ;
module_param( initpos, charp, S_IRUGO );

static int  s_yShift = 0 ;


//---------------------------------------------------------------------------- 
// s1d13u11_tp_open()
// Touch Screen Open
// - uses input core driver
// - dummy method.
//
// param	input	:	input device struct
//
// retval  	0
//   if error, then error code
//
//---------------------------------------------------------------------------- 
static int s1d13u11_tp_open( struct input_dev *input )
{
	DPRINT("(%s-%d)\n",__func__,__LINE__);

	return 0;
}


//---------------------------------------------------------------------------- 
// s1d13u11_tp_release()
// Touch Screen Release
// - uses input core driver
// - dummy method.
//
// param	input	:	input device struct
//
// retval  	0
//   if error, then error code
//
//---------------------------------------------------------------------------- 
static void s1d13u11_tp_release(struct input_dev *input)
{
	DPRINT("(%s-%d)\n",__func__,__LINE__);
	return ;
}

//---------------------------------------------------------------------------- 
// s1d13u11_tp_touch_read()
// Touch Screen Interrupt receive start
// - Touch screen sequence data set
// - Send "CBW" (SPI_SEQUENCE_START) to S1D13U11 Core driver
// - Receive "CSW" response.
//
// param	len	: data length
// param	buf	: data
//
// retval  	0
//   if error, then error code
//
//---------------------------------------------------------------------------- 
unsigned int s1d13u11_tp_touch_read( unsigned short *len, unsigned char *buf )
{
	int	ret =0 ;
	struct s1d13u11_disp_cbw_spi_sequence_start		cbwSeqSt ;
	struct s1d13u11_disp_csw_spi_sequence_start		cswSeqSt ;
	unsigned char * sndData;

	/* touch sequence data set */
	unsigned char seqdata[] = {
				// For y pos
				TP_SEQUENCE_INTDIS,
				TP_SEQUENCE_CSENB,
				TP_SEQUENCE_WRITE,
				0x90,
				TP_SEQUENCE_READ,
				TP_SEQUENCE_READ,
				TP_SEQUENCE_CSDIS,

				// For z1 pos
				TP_SEQUENCE_CSENB,
				TP_SEQUENCE_WRITE,
				0xB0,
				TP_SEQUENCE_READ,
				TP_SEQUENCE_READ,
				TP_SEQUENCE_CSDIS,

				// For z2 pos
				TP_SEQUENCE_CSENB,
				TP_SEQUENCE_WRITE,
				0xC0,
				TP_SEQUENCE_READ,
				TP_SEQUENCE_READ,
				TP_SEQUENCE_CSDIS,

				// For x pos
				TP_SEQUENCE_CSENB,
				TP_SEQUENCE_WRITE,
				0xD0,
				TP_SEQUENCE_READ,
				TP_SEQUENCE_READ,
				TP_SEQUENCE_CSDIS,

				TP_SEQUENCE_INTDNA,
			 };

	DPRINT( "(%s-%d)\n",__func__,__LINE__ );

	DPRINT( "seqdata [%d][%02X %02X %02X %02X %02X|%02X %02X %02X %02X %02X|%02X %02X %02X %02X]\n",
		sizeof(seqdata),
		seqdata[0], seqdata[1], seqdata[2], seqdata[3],
		seqdata[4], seqdata[5], seqdata[6], seqdata[7],
		seqdata[8], seqdata[9],	seqdata[10], seqdata[11],
		seqdata[12], seqdata[13] );


	//----------------------------------------------------------------------
	// Touch Screen Sequence Start
	//----------------------------------------------------------------------
	memset( (void *)&cbwSeqSt, 0x00, sizeof(cbwSeqSt) );
	memset( (void *)&cswSeqSt, 0x00, sizeof(cswSeqSt) );

	// SPI_SEQUENCE_START Message Create
	cbwSeqSt.cbw.cmd	= S1D13U11_DISP_CMD_SPI_SEQUENCE_START ;
	cbwSeqSt.bChannel	= S1D13U11_DISP_SEL_SPI_SEQ_CH0 ;
	cbwSeqSt.bMode		= 0x00 ;
	cbwSeqSt.bCycleTime	= 0x03 ;
	cbwSeqSt.size		= sizeof(seqdata) ;

	DPRINT( " %d %d (%s-%d)\n",(sizeof(struct s1d13u11_disp_cbw_spi_sequence_start) ) , sizeof(seqdata),__func__,__LINE__ );

	// Command Send
	sndData = (unsigned char *)kmalloc(sizeof(seqdata) + sizeof(cbwSeqSt), GFP_KERNEL);
	if( sndData==NULL ) {
		// Send NG
		return( -EFAULT );
	}

	memset( (void *)sndData, 0x00, sizeof(seqdata) + sizeof(cbwSeqSt));
	memcpy( sndData, &cbwSeqSt, sizeof( struct s1d13u11_disp_cbw_spi_sequence_start ));
	memcpy( &sndData[sizeof(struct s1d13u11_disp_cbw_spi_sequence_start)], &seqdata, sizeof( seqdata ));


	ret = s1d13u11_control_cmd_send(
			sizeof(struct s1d13u11_disp_cbw_spi_sequence_start) + sizeof(seqdata),
			sndData,
			sizeof(cswSeqSt),
			(unsigned char *)&cswSeqSt );

	kfree( sndData );

	if( ret!=0 ) {
		// Send NG
		return( -EFAULT );
	}

	// status check
	if( cswSeqSt.csw.stat != S1D13U11_DISP_STAT_SUCCESS ) {
		// Status NG
		return( -EFAULT );
	}

	return 0;
}


//---------------------------------------------------------------------------- 
// s1d13u11_tp_seq_read()
// Touch Screen data read callback processing
// - input data check
// - set input device
// - input device sync
// 
// param	len	: data length
// param	buf	: input data
//
// retval  	0	: SUCCESS
// retval	Not 0	: error code
//
//---------------------------------------------------------------------------- 
unsigned int s1d13u11_tp_seq_read( unsigned short *len, unsigned char *buf )
{
	struct tp_event	*tc;
	unsigned	Rt;
	u16		x ;
	u16		y ;
	u16		z1 ;
	u16		z2 ;
	int		sync = 0 ;

	int		minPos = -1;
	int		maxPos = -1;


	DPRINT("%s START \n", __FUNCTION__);

	/* input data check */
	if( buf==NULL ) {
		return 0 ;
	}

	tc = (struct tp_event *)buf;

	x     = (be16_to_cpu(tc->x) & 0x7FF8 ) >> 3 ;
	y     = (be16_to_cpu(tc->y) & 0x7FF8 ) >> 3 ;
	z1    = (be16_to_cpu(tc->z1) & 0x7FF8 ) >> 3 ;
	z2    = (be16_to_cpu(tc->z2) & 0x7FF8 ) >> 3 ;

	y = abs(y - TP_Y_MAX) ;

	if( y && ((z1 >> 4) !=0) ) {
		/* compute touch pressure resistance using equation #2 */
		Rt = z2;
		Rt -= z1;
		Rt *= x;
		Rt *= 550;
		Rt /= z1;
		Rt = (Rt + 2047) >> 12;
	} else {
		s_tpBuffCnt = 0 ;
		memset( &tpBuff, 0x00, sizeof(tpBuff) );
		Rt = 0;
	}

	/* set input device */
	if ( Rt != 0 && y!=0 && !gPendown) {
		input_report_key(s_indev, BTN_TOUCH, 1);
		DPRINT( "!!!PUSH!!!(%s-%d)\n",__func__,__LINE__ );
		sync = 1 ;
	} else if( gPendown && Rt==0 ){
		input_report_key(s_indev, BTN_TOUCH, 0);

		DPRINT( "!!!RELEASE!!!(%s-%d)\n",__func__,__LINE__ );
		sync = 1 ;
	}else if ( Rt != 0 && y!=0 && gPendown) {
		input_report_key(s_indev, BTN_TOUCH, 1);
		DPRINT( "!!!continue!!!(%s-%d)\n",__func__,__LINE__ );
		sync = 1 ;
	} else {
		DPRINT( "NO SYNC(%s-%d)\n",__func__,__LINE__ );
	}

	if (Rt != 0 && y!=0 && sync==1) {
		DPRINT( "x[%d] y[%d]<%d><%d><%d>(%d-%d) z[%d][%d] Rt[%d] pen[%d](%s-%d)\n", x,y,tpBuff[0].y,tpBuff[1].y,
			tpBuff[2].y,minPos,maxPos,z1,z2,Rt,gPendown,__func__,__LINE__ );
		input_report_abs(s_indev, ABS_X, x );
		input_report_abs(s_indev, ABS_Y, y );
		input_report_abs(s_indev, ABS_PRESSURE, Rt );
		sync = 1 ;
	}

	/* input device sync */
	if( sync==1 ) {
		input_sync( s_indev );
		DPRINT("x=%04x y=%04x Rt=%04x\n", x, y , Rt);
	}

	gPendown = ( Rt != 0 );

	return 0;
}


//---------------------------------------------------------------------------- 
// s1d13u11_tp_atoi()
// Touch Screen data change processing(char->int)
// - Loop input data
// - char to int change
//
// param	name	:	change data
//
// retval  	int data
//
//---------------------------------------------------------------------------- 
static int s1d13u11_tp_atoi( const char *name )
{
	int val = 0;

	for (;; name++) {
		switch (*name) {
			case '0'...'9':
			val = (10 * val) + (*name-'0');
			break;
			default:
			return val;
		}
	}
}


//---------------------------------------------------------------------------- 
// s1d13u11_tp_seq_stop()
// Touch Screen data sequence stop
// - SPI_SEQUENCE_STOP create message
// - command send
//
// param	spich	: spi channel
// 
// retval  	0	: SUCCESS
// retval	Not 0	: error code
//
//---------------------------------------------------------------------------- 
static int s1d13u11_tp_seq_stop( unsigned char spich )
{
	int	ret =0 ;
	struct s1d13u11_disp_cbw_spi_sequence_stop		cbwSeqSt ;
	struct s1d13u11_disp_csw_spi_sequence_stop		cswSeqSt ;

	DPRINT( "(%s-%d)\n",__func__,__LINE__ );

	//----------------------------------------------------------------------
	// Touch Screen Sequence Start
	//----------------------------------------------------------------------
	memset( (void *)&cbwSeqSt, 0x00, sizeof(cbwSeqSt) );
	memset( (void *)&cswSeqSt, 0x00, sizeof(cswSeqSt) );

	// SPI_SEQUENCE_START Message Create
	cbwSeqSt.cbw.cmd    = S1D13U11_DISP_CMD_SPI_SEQUENCE_STOP ;
	cbwSeqSt.bChannel   = spich ;

	// Command Send
	ret = s1d13u11_control_cmd_send(
			sizeof(cbwSeqSt),
			(unsigned char *)&cbwSeqSt,
			sizeof(cswSeqSt),
			(unsigned char *)&cswSeqSt );

	if( ret<sizeof(struct s1d13u11_disp_csw_spi_sequence_stop) ) {
		// Send NG
		DPRINT( "ERROR(%s-%d)\n",__func__,__LINE__ );
		return( -EFAULT );
	}

	// status check
	if( cswSeqSt.csw.stat != S1D13U11_DISP_STAT_SUCCESS ) {
		// Status NG
		DPRINT( "ERROR(%s-%d)\n",__func__,__LINE__ );
		return( -EFAULT );
	}

	return 0 ;
}


//---------------------------------------------------------------------------- 
// s1d13u11_tp_init()
// Touch Screen Initialize
// - Run parameter check
//	 - case None(Input existence)
// 	 - X minimum position set
//   - Y minimum position set
//   - X max position set
//   - Y max position set
// - allocate input device
// - input device setting
// - set touch panel device info
// - set device number
// - initialize spi device(SPI_CONFIG)
// - SPI interrupt set callback
//
// retval  	0
//  if error, then error code
//
//---------------------------------------------------------------------------- 
static int __init s1d13u11_tp_init(void)
{
	int		ret ;
	int		err;

	int		x_min,x_max ;
	int		y_min,y_max ;
	
	struct s1d13u11_disp_cbw_spi_config		cbwCfgData ;
	struct s1d13u11_disp_csw_spi_config		cswCfgData ;

	x_min = TP_X_MIN ;
	y_min = TP_Y_MIN ;
	x_max = TP_X_MAX ;
	y_max = TP_Y_MAX ;
	s_yShift = 0 ;

	//----------------------------------------------------------------------
	// Input Core Driver Setting
	//----------------------------------------------------------------------
	if( initpos != NULL ) {
		s_yShift = s1d13u11_tp_atoi( initpos );

	}
	DPRINT( "s_yShift[%d](%s-%d)\n",s_yShift,__func__,__LINE__ );

	/* allocate input device */
	s_indev = input_allocate_device();

	if (!s_indev) {
		err = -ENOMEM;
		return err ;
	}

	/* input device setting. */
	s_indev->name = S1D13U11_BL_NAME_DRIVER;
	s_indev->phys = "s1d13u11_ts/input0";

	s_indev->evbit[0] = BIT_MASK(EV_KEY) | BIT_MASK(EV_ABS);
	//s_indev->evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
	s_indev->absbit[0] = BIT_MASK(ABS_X) | BIT_MASK(ABS_Y) | BIT_MASK(ABS_PRESSURE);
	//s_indev->absbit[0] = BIT(ABS_X) | BIT(ABS_Y) | BIT(ABS_PRESSURE);	
	s_indev->keybit[BIT_WORD(BTN_TOUCH)] = BIT_MASK(BTN_TOUCH);
	//s_indev->keybit[BITS_TO_LONGS (BTN_TOUCH)] = BIT_MASK(BTN_TOUCH);


	/* set touch panel device info. */
	input_set_abs_params(s_indev, ABS_X,
				x_min, x_max, 0, 0);
	input_set_abs_params(s_indev, ABS_Y,
				y_min, y_max, 0, 0);
	input_set_abs_params(s_indev, ABS_PRESSURE,
				TP_Z_MIN, TP_Z_MAX, 0, 0);

	s_indev->open = s1d13u11_tp_open ;
	s_indev->close = s1d13u11_tp_release ;

	/* set device number */
	err = input_register_device(s_indev);
	if (err) {
		printk("s1d13u11_ts: Failed to register device\n");
		input_free_device(s_indev);
		return err;
	}

	//----------------------------------------------------------------------
	// SPI Device Setting
	//----------------------------------------------------------------------
	memset( (void *)&cbwCfgData, 0x00, sizeof(struct s1d13u11_disp_cbw_spi_config) );
	memset( (void *)&cswCfgData, 0x00, sizeof(struct s1d13u11_disp_csw_spi_config) );

	// SPI_CONFIG Message Create
	cbwCfgData.cbw.cmd = S1D13U11_DISP_CMD_SPI_CONFIG ;

	cbwCfgData.ch0 		= 0xF8 ;
	cbwCfgData.ch0Rate	= 0x04 ;
	cbwCfgData.ch0Mode	= 0x01 ;
	cbwCfgData.ch1 		= 0x30 ;
	cbwCfgData.ch1Rate	= 0x03 ;
	cbwCfgData.ch1Mode	= 0x01 ;

	// Command Send
	ret = s1d13u11_control_cmd_send(
			sizeof(cbwCfgData),
			(unsigned char *)&cbwCfgData,
			sizeof(cswCfgData),
			(unsigned char *)&cswCfgData );

	if( ret < sizeof(struct s1d13u11_disp_csw) ) {
		// Send NG
		DPRINT( "ERROR(%s-%d)\n",__func__,__LINE__ );
		return( -EFAULT );
	}

	// status check
	if( cswCfgData.csw.stat != S1D13U11_DISP_STAT_SUCCESS ) {
		// Status NG
		DPRINT( "ERROR (%s-%d)\n",__func__,__LINE__ );
		return( -EFAULT );
	}

	//----------------------------------------------------------------------
	// register callback
	//----------------------------------------------------------------------
	ret = s1d13u11_recv_int(
			UCHARIS( S1D13U11_DEV_ID_TS ),
			(unsigned int)S1D13U11_EV_TYPE_INT,
			(void *)s1d13u11_tp_touch_read );
	if( ret ) {
		return (int)ret ;
	}

	ret = s1d13u11_recv_int(
			UCHARIS( S1D13U11_DEV_ID_TS ),
			//(unsigned int)S1D13U11_DISP_EVENT_INT_STAT_SPISEQ,
			(unsigned int)S1D13U11_EV_TYPE_SEQ,
			(void *)s1d13u11_tp_seq_read );
	if( ret ) {
		return (int)ret ;
	}

	//----------------------------------------------------------------------
	// Start Touch Screen Sequence
	//----------------------------------------------------------------------
	s1d13u11_tp_touch_read( 0,0 );

	return 0 ;
}

//---------------------------------------------------------------------------- 
// s1d13u11_tp_exit()
// Touch Screen Exit
// - device spi_sequence stop
// - free input device
//
// retval  none
//
//---------------------------------------------------------------------------- 
static void __exit s1d13u11_tp_exit(void)
{

	// Touch Sequence Stop
	s1d13u11_tp_seq_stop( S1D13U11_DISP_SEL_SPI_SEQ_CH0 );

	// free input device
	input_unregister_device( s_indev );
	return ;
}

module_init(s1d13u11_tp_init);
module_exit(s1d13u11_tp_exit);

EXPORT_SYMBOL(s1d13u11_tp_touch_read);
EXPORT_SYMBOL(s1d13u11_tp_seq_read);

MODULE_AUTHOR("USB display project team.");
MODULE_LICENSE("GPL v2");
MODULE_DESCRIPTION("SEIKO EPSON S1D13U11 Touch Screen Driver");
